# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Download Caltech-UCSD Birds 200 dataset.
http://www.vision.caltech.edu/visipedia/CUB-200.html
"""
import os
import tarfile
import tempfile
from glob import glob

import pandas as pd
import wget


def load_data(path=None):
    """Download the CUB200 data set to local storage. This will generate a cub200.csv file, which contains all the path
        information.

    Args:
        path (str, optional): The path to store the CUB200 data. Defaults to None, will save at `tempfile.gettempdir()`.

    Raises:
        FileNotFoundError: When the gernerated CSV file does not match with the extracted dataset.
    """
    if path is None:
        path = os.path.join(tempfile.gettempdir(), ".fe", "CUB200")
    if not os.path.exists(path):
        os.makedirs(path)
    csv_path = os.path.join(path, 'cub200.csv')
    image_compressed_path = os.path.join(path, 'images.tgz')
    annotation_compressed_path = os.path.join(path, 'annotations.tgz')
    image_extracted_path = os.path.join(path, 'images')
    annotation_extracted_path = os.path.join(path, 'annotations-mat')
    if not (os.path.exists(image_compressed_path) and os.path.exists(annotation_compressed_path)):
        print("Downloading data to {} ...".format(path))
        wget.download('http://www.vision.caltech.edu/visipedia-data/CUB-200/images.tgz', path)
        wget.download('http://www.vision.caltech.edu/visipedia-data/CUB-200/annotations.tgz', path)
    if not (os.path.exists(image_extracted_path) and os.path.exists(annotation_extracted_path)):
        print('\nExtracting files ...')
        with tarfile.open(image_compressed_path) as img_tar:
            img_tar.extractall(path)
        with tarfile.open(annotation_compressed_path) as anno_tar:
            anno_tar.extractall(path)
    if not os.path.exists(csv_path):
        img_list = glob(os.path.join(path, 'images', '**', '*.jpg'))
        df = pd.DataFrame(data={'image': img_list})
        df['image'] = df['image'].apply(lambda x: os.path.relpath(x, path))
        df['image'] = df['image'].apply(os.path.normpath)
        df['annotation'] = df['image'].str.replace('images', 'annotations-mat').str.replace('jpg', 'mat')
        df.to_csv(csv_path, index=False)
        print('Data summary is saved at {}'.format(csv_path))
    return csv_path, path



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import tensorflow as tf
import numpy as np
import json
import os


def get_number_of_examples(file_path, show_warning=True, compression=None):
    """
    Returns number of examples in one TFRecord

    Args:
        file_path: Path of TFRecord file
        show_warning: Whether to display warning message
        compression: None, 'GZIP' or 'ZLIB'

    Returns:
        Number of examples in the TFRecord
    """
    _, ext = os.path.splitext(file_path)
    assert "tfrecord" in ext, "please make sure data is in tfrecord format"
    dataset = tf.data.TFRecordDataset(file_path, compression_type=compression)
    example_size = len(next(
        iter(dataset)).numpy()) + 16  # from multiple observations, tfrecord adds 16 byte to each example
    file_size = os.stat(file_path).st_size
    if file_size % example_size != 0 and show_warning:
        print("FastEstimator-Warning: Can't accurately calculate number of examples")
    return max(file_size // example_size, 1)


def get_features(file_path, compression=None):
    """
    Returns the feature information in TFRecords

    Args:
        file_path: Path of TFRecord file
        compression: None, 'GZIP' or 'ZLIB'

    Returns:
        Dictionary containing feature information of TFRecords
    """
    def _get_dtype(ex, feature):
        dtype = list(ex.features.feature[feature].DESCRIPTOR.fields_by_name.keys())
        dtype = np.array(dtype)
        feature_dtype = dtype[[ex.features.feature[feature].HasField(x) for x in dtype]]
        feature_dtype = str(np.squeeze(feature_dtype))
        type_dict = {'bytes_list': tf.string, 'int64_list': tf.int64, 'float_list': tf.float32}
        tf_type = type_dict[feature_dtype]
        return tf_type

    _, ext = os.path.splitext(file_path)
    assert "tfrecord" in ext, "please make sure data is in tfrecord format"
    dataset = tf.data.TFRecordDataset(file_path, compression_type=compression)
    example = tf.train.Example.FromString(next(iter(dataset)).numpy())
    feature_list = list(example.features.feature.keys())
    tf_type_list = [tf.io.FixedLenFeature([], _get_dtype(example, f)) for f in feature_list]
    keys_to_features = dict(zip(feature_list, tf_type_list))
    return keys_to_features


def add_summary(data_dir,
                train_prefix,
                feature_name,
                feature_dtype,
                feature_shape,
                eval_prefix=None,
                num_train_examples=None,
                num_eval_examples=None,
                compression=None):
    """Adds summary.json file to existing path with tfrecords.

    Args:
        data_dir (str): Folder path where tfrecords are stored.
        train_prefix (str): The prefix of all training tfrecord files.
        feature_name (list): Feature name in the tfrecord in a list.
        feature_dtype (list): Original data type for specific feature, this is used for decoding purpose.
        feature_shape (list): Original data shape for specific feature, this is used for reshaping purpose.
        eval_prefix (str, optional): The prefix of all evaluation tfrecord files. Defaults to None.
        num_train_examples (int, optional): The total number of training examples, if None, it will calculate
            automatically. Defaults to None.
        num_eval_examples (int, optional): The total number of validation examples, if None, it will calculate
            automatically. Defaults to None.
        compression (str, optional): None, 'GZIP' or 'ZLIB'. Defaults to None.
    """
    train_files = [f for f in os.listdir(data_dir) if f.startswith(train_prefix)]
    assert len(train_files) > 0, "Couldn't find any training tfrecord files in %s" % data_dir
    dataset = tf.data.TFRecordDataset(os.path.join(data_dir, train_files[0]), compression_type=compression)
    example = tf.train.Example.FromString(next(iter(dataset)).numpy())
    feature_list = list(example.features.feature.keys())
    assert set(feature_list).issuperset(
        set(feature_name)
    ), "feature name should at least be subset of feature name in tfrecords, found %s , given %s." % (str(feature_list),
                                                                                                      str(feature_name))
    if not num_train_examples:
        # num_train_examples = [get_number_of_examples(os.path.join(data_dir, f)) for f in train_files]
        num_trian_files = len(train_files)
        logging_interval = max(num_trian_files // 10, 1)
        num_train_examples = []
        for i in range(num_trian_files):
            if (i + 1) % logging_interval == 0:
                print("FastEstimator: Calculating number of examples for train %d/%d" % (i + 1, num_trian_files))
            num_train_examples.append(
                get_number_of_examples(file_path=os.path.join(data_dir, train_files[i]),
                                       show_warning=i == 0,
                                       compression=compression))
    summary = {
        "feature_name": feature_name,
        "feature_dtype": feature_dtype,
        "feature_shape": feature_shape,
        "train_files": train_files,
        "num_train_examples": num_train_examples
    }
    if eval_prefix:
        eval_files = [f for f in os.listdir(data_dir) if f.startswith(eval_prefix)]
        assert len(eval_files) > 0, "Couldn't find any validation tfrecord files in %s" % data_dir
        if not num_eval_examples:
            # num_eval_examples = [get_number_of_examples(os.path.join(data_dir, f)) for f in eval_files]
            num_eval_files = len(eval_files)
            logging_interval = max(num_eval_files // 10, 1)
            num_eval_examples = []
            for i in range(num_eval_files):
                if (i + 1) % logging_interval == 0:
                    print("FastEstimator: Calculating number of examples for eval %d/%d" % (i + 1, num_eval_files))
                num_eval_examples.append(
                    get_number_of_examples(file_path=os.path.join(data_dir, eval_files[i]),
                                           show_warning=i == 0,
                                           compression=compression))
        summary["eval_files"] = eval_files
        summary["num_eval_examples"] = num_eval_examples
    if compression:
        summary["compression"] = compression
    with open(os.path.join(data_dir, "summary0.json"), 'w') as fp:
        json.dump(summary, fp, indent=4)
    print("FastEstimator: Writing " + str(os.path.join(data_dir, "summary0.json")))



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from collections import ChainMap

import tensorflow as tf
from tensorflow.python.framework import ops as tfops

from fastestimator.network.model import ModelOp
from fastestimator.util.op import get_inputs_by_op, get_op_from_mode, verify_ops, write_outputs_by_key
from fastestimator.util.schedule import Scheduler
from fastestimator.util.util import NonContext, flatten_list


class Network:
    def __init__(self, ops):
        if not isinstance(ops, list):
            ops = [ops]
        self.ops = ops
        self.model_schedule = {}
        self.op_schedule = {}
        self.current_epoch_ops = {}
        self.model = {}
        self.all_losses = []
        self.epoch_losses = []
        self.all_output_keys = set()
        self.stop_training = False
        self.num_devices = 1

    def prepare(self, mode_list, distribute_strategy):
        all_output_keys = []
        for mode in mode_list:
            signature_epoch, mode_ops = self._get_signature_epoch(mode)
            epoch_ops_map = {}
            epoch_model_map = {}
            for epoch in signature_epoch:
                epoch_ops = []
                epoch_model = []
                # generate ops for specific mode and epoch
                for op in mode_ops:
                    if isinstance(op, Scheduler):
                        scheduled_op = op.get_current_value(epoch)
                        if scheduled_op:
                            epoch_ops.append(scheduled_op)
                    else:
                        epoch_ops.append(op)
                # check the ops
                verify_ops(epoch_ops, "Network")
                # create model list
                for op in epoch_ops:
                    all_output_keys.append(op.outputs)
                    if isinstance(op, ModelOp):
                        if op.model.keras_model is None:
                            with distribute_strategy.scope() if distribute_strategy else NonContext():
                                op.model.keras_model = op.model.model_def()
                                op.model.keras_model.optimizer = op.model.optimizer
                                op.model.keras_model.loss_name = op.model.loss_name
                                op.model.keras_model.model_name = op.model.model_name
                                assert op.model.model_name not in self.model, \
                                    "duplicated model name: {}".format(op.model.model_name)
                                self.model[op.model.model_name] = op.model.keras_model
                                if op.model.loss_name not in self.all_losses:
                                    self.all_losses.append(op.model.loss_name)
                        if op.model.keras_model not in epoch_model:
                            epoch_model.append(op.model.keras_model)
                assert epoch_model, "Network has no model for epoch {}".format(epoch)
                epoch_ops_map[epoch] = epoch_ops
                epoch_model_map[epoch] = epoch_model
            self.op_schedule[mode] = Scheduler(epoch_dict=epoch_ops_map)
            self.model_schedule[mode] = Scheduler(epoch_dict=epoch_model_map)
        self.all_output_keys = set(flatten_list(all_output_keys)) - {None}

    def _get_signature_epoch(self, mode):
        signature_epoch = [0]
        mode_ops = get_op_from_mode(self.ops, mode)
        for op in mode_ops:
            if isinstance(op, Scheduler):
                signature_epoch.extend(op.keys)
        return list(set(signature_epoch)), mode_ops

    def load_epoch(self, epoch, mode):
        ops = self.op_schedule[mode].get_current_value(epoch)
        model_list = self.model_schedule[mode].get_current_value(epoch)
        epoch_losses = []
        for model in model_list:
            if model.loss_name not in epoch_losses:
                epoch_losses.append(model.loss_name)
        self.epoch_losses = epoch_losses
        return ops, model_list, epoch_losses

    def run_step(self, batch, ops, model_list, epoch_losses, state, warm_up=False):
        prediction = {}
        batch = ChainMap(prediction, batch)
        mode = state["mode"]
        global_batch_size = state["batch_size"]
        num_model = len(model_list)
        # use gradient tape for train, otherwise use a dummy tape
        with tf.GradientTape(persistent=True) if mode == "train" else NonContext() as tape:
            state['tape'] = tape
            self._forward(batch, state, ops)
            reduced_loss = self._reduce_loss(batch, global_batch_size, epoch_losses, warm_up)
        # update model only for train mode
        if mode == "train":
            for idx in range(num_model):
                model = model_list[idx]
                loss = reduced_loss[model.loss_name]
                optimizer = model.optimizer
                if warm_up:
                    with tfops.init_scope():  # pylint: disable=not-context-manager
                        _ = optimizer.iterations
                        optimizer._create_hypers()  # pylint: disable=protected-access
                        optimizer._create_slots(model_list[idx].trainable_variables)  # pylint: disable=protected-access
                else:
                    gradients = tape.gradient(loss, model.trainable_variables)
                    optimizer.apply_gradients(zip(gradients, model.trainable_variables))
        del state['tape']
        del tape
        return prediction

    @staticmethod
    def _forward(batch, state, ops):
        data = None
        for op in ops:
            data = get_inputs_by_op(op, batch, data)
            data = op.forward(data, state)
            if op.outputs:
                write_outputs_by_key(batch, data, op.outputs)

    def _reduce_loss(self, batch, global_batch_size, epoch_losses, warm_up):
        reduced_loss = {}
        for loss_name in epoch_losses:
            element_wise_loss = batch[loss_name]
            if warm_up:
                assert element_wise_loss.shape[0] == global_batch_size / self.num_devices, \
                    "please make sure loss is element-wise loss"
            reduced_loss[loss_name] = tf.reduce_sum(element_wise_loss) / global_batch_size
        return reduced_loss



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import types

import tensorflow as tf

from fastestimator.util.op import TensorOp


class FEModel:
    def __init__(self, model_def, model_name, optimizer, loss_name="loss"):
        assert isinstance(model_def, types.FunctionType), \
            "must provide function definition or lambda function as model_def"
        assert isinstance(loss_name, str), "loss must be a string key of loss tensor"
        self.model_def = model_def
        self.loss_name = loss_name
        self.optimizer = optimizer
        self.model_name = model_name
        self._check_optimizer()
        self.keras_model = None

    def _check_optimizer(self):
        if isinstance(self.optimizer, str):
            optimizer_fn = {
                'adadelta': tf.optimizers.Adadelta,
                'adagrad': tf.optimizers.Adagrad,
                'adam': tf.optimizers.Adam,
                'adamax': tf.optimizers.Adamax,
                'nadam': tf.optimizers.Nadam,
                'rmsprop': tf.optimizers.RMSprop,
                'sgd': tf.optimizers.SGD
            }
            self.optimizer = optimizer_fn[self.optimizer]()
        else:
            assert isinstance(self.optimizer, tf.optimizers.Optimizer), \
                "must provide provide must provide tf.optimizer.Optimizer instance as optimizer"


class ModelOp(TensorOp):
    def __init__(self, model, inputs=None, outputs=None, mode=None, track_input=False):
        super().__init__(inputs=inputs, outputs=outputs, mode=mode)
        assert isinstance(model, FEModel), "must provide a FEModel in as input"
        self.model = model
        self.track_input = track_input

    def forward(self, data, state):
        training = state['mode'] == "train"
        if self.track_input and training:
            tape = state['tape']
            tape.watch(data)
        data = self.model.keras_model(data, training=training)
        return data



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import tensorflow as tf
from tensorflow.python.keras.losses import Loss as tfLoss

from fastestimator.util.op import TensorOp


class Loss(TensorOp):
    """
    A base class for loss operations. It can be used directly to perform value pass-through (see the adversarial
    training showcase for an example of when this is useful)
    """
    def __init__(self, inputs=None, outputs="loss", mode=None):
        super().__init__(inputs=inputs, outputs=outputs, mode=mode)


class MeanSquaredError(Loss):
    def __init__(self, y_true=None, y_pred=None, inputs=None, outputs="loss", mode=None, **kwargs):
        """Calculate mean squared error loss, the rest of the keyword argument will be passed to
           tf.losses.MeanSquaredError

           Args:
            y_true: ground truth label key
            y_pred: prediction label key
            inputs: A tuple or list like: [<y_true>, <y_pred>]
            outputs: Where to store the computed loss value (not required under normal use cases)
            mode: 'train', 'eval', 'test', or None
            kwargs: Arguments to be passed along to the tf.losses constructor
           """
        if 'reduction' in kwargs:
            raise KeyError("parameter 'reduction' not allowed")
        inputs = validate_loss_inputs(inputs, y_true, y_pred)
        super().__init__(inputs=inputs, outputs=outputs, mode=mode)
        self.loss_obj = tf.losses.MeanSquaredError(reduction='none', **kwargs)

    def forward(self, data, state):
        true, pred = data
        return self.loss_obj(true, pred)


class SparseCategoricalCrossentropy(Loss):
    def __init__(self, y_true=None, y_pred=None, inputs=None, outputs="loss", mode=None, **kwargs):
        """Calculate sparse categorical cross entropy, the rest of the keyword argument will be passed to
           tf.losses.SparseCategoricalCrossentropy

        Args:
            y_true: ground truth label key
            y_pred: prediction label key
            inputs: A tuple or list like: [<y_true>, <y_pred>]
            outputs: Where to store the computed loss value (not required under normal use cases)
            mode: 'train', 'eval', 'test', or None
            kwargs: Arguments to be passed along to the tf.losses constructor. Passing the 'reduction' arg will raise a
                    KeyError
        """
        if 'reduction' in kwargs:
            raise KeyError("parameter 'reduction' not allowed")
        inputs = validate_loss_inputs(inputs, y_true, y_pred)
        super().__init__(inputs=inputs, outputs=outputs, mode=mode)
        self.loss_obj = tf.losses.SparseCategoricalCrossentropy(reduction='none', **kwargs)

    def forward(self, data, state):
        y_true, y_pred = data
        return self.loss_obj(y_true, y_pred)


class BinaryCrossentropy(Loss):
    def __init__(self, y_true=None, y_pred=None, inputs=None, outputs="loss", mode=None, **kwargs):
        """Calculate binary cross entropy, the rest of the keyword argument will be passed to
                  tf.losses.BinaryCrossentropy

       Args:
           y_true: ground truth label key
           y_pred: prediction label key
           inputs: A tuple or list like: [<y_true>, <y_pred>]
           outputs: Where to store the computed loss value (not required under normal use cases)
           mode: 'train', 'eval', 'test', or None
           kwargs: Arguments to be passed along to the tf.losses constructor
       """
        if 'reduction' in kwargs:
            raise KeyError("parameter 'reduction' not allowed")
        inputs = validate_loss_inputs(inputs, y_true, y_pred)
        super().__init__(inputs=inputs, outputs=outputs, mode=mode)
        self.loss_obj = tf.losses.BinaryCrossentropy(reduction='none', **kwargs)

    def forward(self, data, state):
        true, pred = data
        loss = self.loss_obj(true, pred)
        loss_dim = len(loss.shape)
        if loss_dim > 1:
            loss = tf.reduce_mean(loss, axis=list(range(1, loss_dim)))
        return loss


class MixUpLoss(Loss):
    """
    This class should be used in conjunction with MixUpBatch to perform mix-up training, which helps to reduce
    over-fitting, stabilize GAN training, and harden against adversarial attacks (https://arxiv.org/abs/1710.09412)
    """
    def __init__(self, loss, lam=None, y_true=None, y_pred=None, inputs=None, outputs="loss", mode=None):
        """
        Args:
            loss (func): A loss object (tf.losses) which can be invoked like "loss(true, pred)". It's reduction method
                        will be overridden to 'none'
            lam: The key of the lambda value generated by MixUpBatch
            y_true: ground truth label key
            y_pred: prediction label key
            inputs: A tuple or list like: [<lam>, <y_true>, <y_pred>]
            outputs: Where to store the computed loss value (not required under normal use cases)
            mode: 'train', 'eval', 'test', or None
        """
        assert isinstance(loss, tfLoss), "MixUpLoss requires a TensorFlow loss function"
        loss_config = loss.get_config()
        loss_config['reduction'] = 'none'
        inputs = validate_loss_inputs(inputs, lam, y_true, y_pred)
        super().__init__(inputs=inputs, outputs=outputs, mode=mode)
        self.loss_obj = loss.from_config(loss_config)

    def forward(self, data, state):
        lam, true, pred = data
        loss1 = self.loss_obj(true, pred)
        loss2 = self.loss_obj(tf.roll(true, shift=1, axis=0), pred)
        return lam * loss1 + (1.0 - lam) * loss2


def validate_loss_inputs(inputs, *args):
    """
    A method to ensure that either the inputs array or individual input arguments are specified, but not both
    Args:
        inputs: None or a tuple/list of arguments
        *args: a tuple of arguments or Nones
    Returns:
        either 'inputs' or the args tuple depending on which is populated
    """
    if inputs is None:  # Using args
        assert all(map(lambda x: x is not None, args)), \
            "If the 'inputs' field is not provided then all individual input arguments must be specified"
        inputs = args
    else:  # Using Inputs
        assert all(map(lambda x: x is None, args)), \
            "If the 'inputs' field is provided then individual input arguments may not be specified"
        assert len(inputs) == len(args), \
            "{} inputs were provided, but {} were required".format(len(inputs), len(args))
    return inputs



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import matplotlib.pyplot as plt
import numpy as np
import tensorflow as tf
import tensorflow_probability as tfp


def show_text(axis, background, text, title=None):
    """
    Plots a given image onto an axis

    Args:
        axis: The matplotlib axis to plot on
        background: A background image to display behind the text (useful for sizing the plot correctly)
        text: The text to display
        title: A title for the image
    """
    # matplotlib doesn't support (x,y,1) images, so convert them to (x,y)
    if len(background.shape) == 3 and background.shape[2] == 1:
        background = np.reshape(background, (background.shape[0], background.shape[1]))

    axis.axis('off')
    axis.imshow(background, cmap=plt.get_cmap(name="Greys_r"), vmin=0, vmax=1)
    axis.text(0.5, 0.5, text, ha='center', va='center', transform=axis.transAxes, wrap=False, family='monospace')
    if title is not None:
        axis.set_title(title)


def show_image(axis, im, title=None):
    """
    Plots a given image onto an axis

    Args:
        axis: The matplotlib axis to plot on
        im: The image to display (width X height)
        title: A title for the image
    """
    if axis is None:
        fig, axis = plt.subplots(1, 1)
    axis.axis('off')
    im = np.asarray(im)
    if np.max(im) <= 1 and np.min(im) >= 0:  # im is [0,1]
        im = (im * 255).astype(np.uint8)
    elif np.max(im) <= 1:  # im is [-1, 1]
        im = ((im + 1) * 127.5).astype(np.uint8)
    else:  # im is already 255
        im = im.astype(np.uint8)
    # matplotlib doesn't support (x,y,1) images, so convert them to (x,y)
    if len(im.shape) == 3 and im.shape[2] == 1:
        im = np.reshape(im, (im.shape[0], im.shape[1]))
    axis.imshow(im)
    if title is not None:
        axis.set_title(title)


def show_gray_image(axis, im, title=None, color_map="inferno"):
    """
    Plots a given image onto an axis

    Args:
        axis: The matplotlib axis to plot on
        im: The image to display (width X height)
        title: A title for the image
        color_map: The color set to be used (since the image is gray scale)
    """
    axis.axis('off')
    axis.imshow(im, cmap=plt.get_cmap(name=color_map), vmin=0, vmax=1)
    if title is not None:
        axis.set_title(title)


@tf.function
def gaussian_kernel(mean, std):
    """
    Args:
        mean: The center of the normal distribution
        std: The standard deviation of the normal distribution
    Returns:
        A gaussian kernel defined by the specified normal distribution, expanded for multiplication against a
        3-channel image
    """
    # kernel size should be 6x the std in order to be similar to global op
    size = tf.maximum(2.0, tf.math.ceil(3 * std))
    d = tfp.distributions.Normal(mean, std)
    vals = d.prob(tf.range(start=-size, limit=size + 1, dtype=tf.float32))
    kernel = tf.einsum('i,j->ij', vals, vals)
    kernel = kernel / tf.reduce_sum(kernel)
    # Convert the kernel to 3 channel TODO make dynamic
    kernel = tf.stack([kernel, kernel, kernel], axis=2)
    # Expand kernel shape to match conv2d spec
    kernel = kernel[:, :, :, tf.newaxis]
    return kernel


@tf.function
def blur_image(image, std):
    """
    Args:
        image: A 3 channel input image
        std: The standard deviation of the gaussian filter to be applied
    Returns:
        'image' blurred by a gaussian filter
    """
    kernel = gaussian_kernel(0, std)
    return tf.nn.depthwise_conv2d(image, kernel, strides=[1, 1, 1, 1], padding="SAME")


@tf.function
def blur_image_fft(image, std):
    """
    This method is equivalent to blur_image(), but is slower right now due to a bug in tensorflow related to graph
    construction for the irfft2d function. TODO re-evaluate performance trade-off as new version of TF2 are released
    Args:
        image: An n-channel input image
        std: The standard deviation of the gaussian filter to be applied
    Returns:
        'image' blurred by a gaussian filter
    """
    mean = 0
    # Wiki article on gaussian blurring indicates that a 6s * 6s filter is basically equivalent to global kernel
    half_kernel_size = tf.maximum(2.0, tf.math.ceil(3.0 * std))

    kernel_size = 2 * half_kernel_size + 1  # Kernel size will always be odd
    b, h, w, c = image.shape
    # Convert to batch-channel-h-w
    tf_image = tf.transpose(image, [0, 3, 1, 2])
    fft_length = [h, w]
    # Need to pad the fft size based on the kernel size. Also, the last axis needs to be padded by one if it ends up
    # being odd. The extra pixel will get cropped off at the end
    fft_length_padded = [fft_length[0] + kernel_size, fft_length[1] + kernel_size + (1 + w) % 2]
    ft_image = tf.signal.rfft2d(tf_image, fft_length=fft_length_padded)

    d = tfp.distributions.Normal(mean, std)
    vals = d.prob(tf.range(start=-half_kernel_size, limit=half_kernel_size + 1, dtype=tf.float32))
    kernel = tf.einsum('i,j->ij', vals, vals)
    kernel = kernel / tf.reduce_sum(kernel)

    big_kernel = tf.eye(num_rows=kernel_size, num_columns=kernel_size, batch_shape=[b, c], dtype=kernel.dtype)
    kernel = tf.matmul(big_kernel, kernel)
    ft_kernel = tf.signal.rfft2d(kernel, fft_length=fft_length_padded)

    ft_image = ft_image * ft_kernel

    tf_image = tf.signal.irfft2d(ft_image)
    # Convert to batch-dim1-dim2-channel
    tf_image = tf.transpose(tf_image, [0, 2, 3, 1])
    # Cast to int for indexing
    half_kernel_size = tf.cast(half_kernel_size, dtype=tf.int32)
    # Crop the extra pixels
    tf_image = tf_image[:, half_kernel_size:h + half_kernel_size, half_kernel_size:w + half_kernel_size, :]
    return tf_image


def rfft2d_freqs(h, w):
    """
    Args:
        h: height of an image
        w: width of an image
    Returns:
        A 2d fourier spectrum corresponding to the given image dimensions
    """
    fy = np.fft.fftfreq(h)[:, None]
    # when we have an odd input dimension we need to keep one additional
    # frequency and later cut off 1 pixel
    fx = np.fft.fftfreq(w)[:w // 2 + 1 + w % 2]
    return np.sqrt(fx * fx + fy * fy)


@tf.function
def fft_vars_to_whitened_im(fft_vars, scale, width):
    """
    Args:
        fft_vars: A tensor of dimension: real/imaginary X batch X channel X height X width
        scale: A matrix which normalizes the energy of an image
        width: The true width of the pixel image (needed to resolve ambiguity introduced by the fourier transform)
    Returns:
        A pixel-space image corresponding to the given input variables, after applying a normalization by 'scale'
    """
    spectrum = tf.complex(fft_vars[0], fft_vars[1])
    spectrum = scale * spectrum  # whiten the fft domain image (de-correlates adjacent pixels)
    image = tf.signal.irfft2d(spectrum)
    image = tf.transpose(image, (0, 2, 3, 1))  # fft domain required width and height to be the last dimensions
    image = image[:, :, :width, :]  # Crop the extra pixel that results if width is odd
    image = image / 4.0  # Make the image develop more slowly / controlled. Not sure why 4 vs any other factor
    return image


@tf.function
def fft_vars_to_im(fft_vars, width):
    """
        Args:
            fft_vars: A tensor of dimension: real/imaginary X batch X channel X height X width
            width: The true width of the pixel image (needed to resolve ambiguity introduced by the fourier transform)
        Returns:
            A pixel-space image corresponding to the given input variables
        """
    spectrum = tf.complex(fft_vars[0], fft_vars[1])
    image = tf.signal.irfft2d(spectrum)
    image = tf.transpose(image, (0, 2, 3, 1))
    image = image[:, :, :width, :]
    image = image / 4.0  # Make the image develop more slowly / controlled. Not sure why 4 vs any other factor
    return image



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import tensorflow as tf


class SaliencyMask(object):
    """Base class for saliency masks. Alone, this class doesn't do anything.
    This code was adapted from https://github.com/PAIR-code/saliency to be compatible with TensorFlow 2.0
    """
    def __init__(self, model):
        """
        Args:
            model: The ML model to evaluate masks on
        """
        self.model = model

    @tf.function
    def get_mask(self, model_input):
        """
        Args:
            model_input: Input tensor, shaped for the model ex. (1, 299, 299, 3)
        Returns:
            A saliency mask
        """
        raise NotImplementedError('A derived class should implemented GetMask()')

    def get_smoothed_mask(self, model_input, stdev_spread=.15, nsamples=25, magnitude=True, **kwargs):
        """
        Args:
            model_input: Input tensor, shaped for the model ex. (1, 299, 299, 3)
            stdev_spread: Amount of noise to add to the input, as fraction of the
                        total spread (x_max - x_min). Defaults to 15%.
            nsamples: Number of samples to average across to get the smooth gradient.
            magnitude: If true, computes the sum of squares of gradients instead of
                     just the sum. Defaults to true.
        Returns:
            A saliency mask that is smoothed with the SmoothGrad method
        """
        stdev = stdev_spread * (tf.reduce_max(model_input) - tf.reduce_min(model_input))

        # Adding noise to the image might cause the max likelihood class value to change, so need to keep track of
        # which class we're comparing to
        initial_predictions = self.model(model_input)
        class_indices = tf.reshape(tf.argmax(initial_predictions, 1, output_type='int64'), (model_input.shape[0], 1))
        row_indices = tf.reshape(tf.range(class_indices.shape[0], dtype='int64'), (class_indices.shape[0], 1))
        classes = tf.concat([row_indices, class_indices], 1)

        total_gradients = tf.zeros_like(model_input, dtype='float32')
        for _ in tf.range(nsamples):
            noise = tf.random.normal(model_input.shape, mean=0, stddev=stdev)
            x_plus_noise = model_input + noise
            grad = self.get_mask(x_plus_noise, classes=classes, **kwargs)
            if magnitude:
                total_gradients = tf.add(tf.multiply(grad, grad), total_gradients)
            else:
                total_gradients = tf.add(grad, total_gradients)

        return total_gradients / nsamples


class GradientSaliency(SaliencyMask):
    """A SaliencyMask class that computes saliency masks with a gradient."""
    def __init__(self, model):
        super().__init__(model)

    @tf.function
    def get_mask(self, model_input, classes=None):
        """
        Args:
            model_input: Input tensor, shaped for the model ex. (1, 299, 299, 3)
            classes: A tensor describing which output index to consider for each element in the input batch.
                    ex. [[0, 5], [1, 7]] for an input with batch size 2 where indexes 5 and 7 are the targets
        Returns:
            A vanilla gradient mask
        """
        with tf.GradientTape() as tape:
            tape.watch(model_input)
            logits = self.model(model_input, training=False)
            if classes is None:
                y = tf.reduce_max(logits, 1)  # The maximum likelihood class scores
            else:
                y = tf.gather_nd(logits, classes)
        return tape.gradient(y, model_input)


class IntegratedGradients(GradientSaliency):
    """A SaliencyMask class that implements the integrated gradients method.

    https://arxiv.org/abs/1703.01365
    """
    @tf.function
    def get_mask(self, model_input, input_baseline=None, steps=25, classes=None):
        """
        Args:
            model_input: Input tensor, shaped for the model ex. (1, 299, 299, 3)
            input_baseline: Baseline value used in integration. Defaults to 0.
            steps: Number of integrated steps between baseline and image.
            classes: A tensor describing which output index to consider for each element in the input batch.
                    ex. [[0, 5], [1, 7]] for an input with batch size 2 where indexes 5 and 7 are the targets
        Returns:
            An integrated gradients mask
        """
        if input_baseline is None:
            input_baseline = tf.zeros_like(model_input, dtype=model_input.dtype)

        assert input_baseline.shape == model_input.shape

        x_diff = model_input - input_baseline

        total_gradients = tf.zeros_like(model_input, dtype='float32')

        # Performing this integration might cause the max likelihood class value to change, so need to keep track of
        # which class we're comparing to
        if classes is None:
            initial_predictions = self.model(model_input)
            class_indices = tf.reshape(tf.argmax(initial_predictions, 1, output_type='int64'),
                                       (model_input.shape[0], 1))
            row_indices = tf.reshape(tf.range(class_indices.shape[0], dtype='int64'), (class_indices.shape[0], 1))
            classes = tf.concat([row_indices, class_indices], 1)

        for alpha in tf.linspace(0.0, 1.0, steps):
            x_step = input_baseline + alpha * x_diff
            total_gradients += super().get_mask(x_step, classes=classes)

        return tf.multiply(total_gradients, x_diff)



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import math
import os

import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
# noinspection PyPackageRequirements
import tensorflow as tf
# noinspection PyPackageRequirements
import umap
from matplotlib.lines import Line2D
from mpl_toolkits.mplot3d import Axes3D
# noinspection PyPackageRequirements
from tensorflow.python import keras
from tqdm import tqdm, trange

from fastestimator.util.loader import ImageLoader
from fastestimator.util.util import load_dict, Suppressor

assert Axes3D  # Axes3D is used to enable projection='3d', but will show up as unused without the assert


def map_classes_to_colors(classifications):
    classes = set(classifications)
    num_classes = len(classes)
    colors = sns.hls_palette(n_colors=num_classes, s=0.95) if num_classes > 10 else sns.color_palette("colorblind")
    class_to_color = {clazz: idx for idx, clazz in enumerate(classes)}
    return [colors[class_to_color[clazz]]
            for clazz in classifications], {clazz: colors[class_to_color[clazz]]
                                            for clazz in classes}


def draw_umaps(layer_outputs, classifications, dictionary=None, layers=None, layer_ids=None, save=False, save_path='.',
               legend_mode='shared'):
    color_list, color_map = map_classes_to_colors(classifications)

    num_layers = len(layer_outputs)
    n_components = len(layer_outputs[0][0])
    num_cols = math.ceil(math.sqrt(num_layers))
    num_rows = math.ceil(num_layers / num_cols)
    layer_grid_location = {idx: (idx // num_cols, idx % num_cols) for idx in range(num_layers)}

    sns.set_context('paper')
    if n_components == 3:
        fig, axs = plt.subplots(num_rows, num_cols, figsize=(4 * num_cols, 2.8 * num_rows),
                                subplot_kw={'projection': '3d'})
    else:
        fig, axs = plt.subplots(num_rows, num_cols, figsize=(4 * num_cols, 2.8 * num_rows))

    # If only one row, need to re-format the axs object for consistency. Likewise for columns
    if num_rows == 1:
        axs = [axs]
        if num_cols == 1:
            axs = [axs]

    # some of the columns in the last row might be unused, so disable them
    last_column_idx = num_cols - (num_rows * num_cols - num_layers) - 1
    for i in range(last_column_idx + 1, num_cols):
        axs[num_rows - 1][i].axis('off')

    # Turn off axis since numeric values are not meaningful
    for i in range(num_rows):
        for j in range(num_cols):
            axis = axs[i][j]
            axis.set_yticks([], [])
            axis.set_yticklabels([])
            axis.set_xticks([], [])
            axis.set_xticklabels([])

    for idx, layer in enumerate(layer_outputs):
        ax = axs[layer_grid_location[idx][0]][layer_grid_location[idx][1]]
        layer_id = idx if layer_ids is None else layer_ids[idx]
        layer_name = "" if layers is None else ": " + layers[layer_id].name
        ax.set_title("Layer {}{}".format(layer_id, layer_name))
        if n_components == 1:
            ax.scatter(layer[:, 0], range(len(layer)), c=color_list, s=3)
        if n_components == 2:
            ax.scatter(layer[:, 0], layer[:, 1], c=color_list, s=3)
        if n_components == 3:
            ax.scatter(layer[:, 0], layer[:, 1], layer[:, 2], c=color_list, s=3)
    plt.tight_layout()

    if legend_mode != 'off':
        legend_elements = [
            Line2D([0], [0], marker='o', color='w', markerfacecolor=color_map[clazz],
                   label=clazz if dictionary is None else dictionary[clazz], markersize=7) for clazz in color_map
        ]
        if legend_mode == 'shared' and num_rows > 1:
            if last_column_idx == num_cols - 1:
                fig.subplots_adjust(bottom=0.15)
                fig.legend(handles=legend_elements, loc='lower center', ncol=num_cols + 1)
            else:
                axs[num_rows - 1][last_column_idx + 1].legend(handles=legend_elements, loc='center', fontsize='large')
        else:
            for i in range(num_rows):
                for j in range(num_cols):
                    if i == num_rows - 1 and j > last_column_idx:
                        break
                    axs[i][j].legend(handles=legend_elements, loc='best', fontsize='small')

    if not save:
        plt.show()
    else:
        if save_path is None or save_path == "":
            save_path = "."
        os.makedirs(save_path, exist_ok=True)
        save_file = os.path.join(save_path, 'umaps.png')
        print("Saving to {}".format(save_file))
        plt.savefig(save_file, dpi=300)


class Evaluator(object):
    def __init__(self, model, layers=None):
        """
        Args:
            model: The ML model to generate outputs from
            layers: The layer indices to be investigated by the evaluator. If not provided then all layers will be used
        """
        self.model = model
        self.layers = layers if layers is not None and len(layers) > 0 else [i for i in range(model.layers)]
        self.num_layers = len(self.layers)
        self.functor = tf.keras.backend.function([self.model.input], [self.model.layers[i].output for i in self.layers])

    def evaluate(self, x):
        layer_outputs = self.functor([x])
        return [np.reshape(layer, [x.shape[0], -1]) for layer in layer_outputs]


class FileCache(object):
    def __init__(self, root_path, layers, umap_parameters):
        self.root_path = root_path
        print("Saving cache files to {}".format(self.root_path))
        os.makedirs(self.root_path, exist_ok=True)
        self.idx = 0
        self.layers = layers
        self.num_layers = len(self.layers)
        self.fit = umap.UMAP(**umap_parameters)

    def save(self, data, classes):
        if len(data) != self.num_layers:
            raise IndexError("Inconsistent Layer Count Detected")
        [
            np.save(os.path.join(self.root_path, "layer{}-batch{}.npy".format(self.layers[layer], self.idx)),
                    data[layer]) for layer in range(self.num_layers)
        ]
        np.save(os.path.join(self.root_path, "class{}.npy".format(self.idx)), classes)
        self.idx += 1

    def batch_cached(self, batch_id):
        return os.path.isfile(os.path.join(self.root_path, "class{}.npy".format(batch_id))) and all([
            os.path.isfile(os.path.join(self.root_path, "layer{}-batch{}.npy".format(self.layers[layer], batch_id)))
            for layer in range(self.num_layers)
        ])

    def load_and_transform(self, batches=None):
        if batches is None:
            batches = self.idx
        data = [None for _ in range(self.num_layers)]
        classes = []

        if batches == 0:
            return data, classes

        for layer in trange(self.num_layers, desc='Computing UMaps', unit='layer'):
            layer_data = []
            for batch in trange(batches, desc='Loading Cache', unit='batch', leave=False):
                dat = np.load(os.path.join(self.root_path, "layer{}-batch{}.npy".format(self.layers[layer], batch)),
                              allow_pickle=True)
                layer_data.append(dat)
            layer_data = np.concatenate(layer_data, axis=0)
            with Suppressor():  # Silence a bunch of numba warnings
                data[layer] = self.fit.fit_transform(layer_data)

        for batch in range(batches):
            clazz = np.load(os.path.join(self.root_path, "class{}.npy".format(batch)), allow_pickle=True)
            classes.extend(clazz)

        return data, classes


def umap_layers(model_path, input_root_path, print_layers=False, strip_alpha=False, layers=None, input_extension=None,
                batch=10, use_cache=True, cache_dir=None, dictionary_path=None, save=False, save_dir=None,
                legend_mode='shared', umap_parameters=None):
    if umap_parameters is None:
        umap_parameters = {}
    if save_dir is None:
        save_dir = os.path.dirname(model_path)
    if cache_dir is None:
        # If the user passes the input dir as a relative path without ./ then dirname will contain all path info
        if os.path.basename(input_root_path) == "":
            cache_dir = os.path.dirname(input_root_path) + "__layer_outputs"
        else:
            cache_dir = os.path.join(os.path.dirname(input_root_path),
                                     os.path.basename(input_root_path) + "__layer_outputs")

    network = keras.models.load_model(model_path)
    if print_layers:
        for idx, layer in enumerate(network.layers):
            print("{}: {} --- output shape: {}".format(idx, layer.name, layer.output_shape))
        return

    evaluator = Evaluator(network, layers=layers)
    loader = ImageLoader(input_root_path, network, batch=batch, input_extension=input_extension,
                         strip_alpha=strip_alpha)
    cache = FileCache(cache_dir, evaluator.layers, umap_parameters) if use_cache else None

    classes = []
    layer_outputs = None
    for batch_id, (batch_inputs, batch_classes) in enumerate(tqdm(loader, desc='Computing Outputs', unit='batch')):
        if use_cache and cache.batch_cached(batch_id):
            continue
        batch_layer_outputs = evaluator.evaluate(batch_inputs)
        if use_cache:
            cache.save(batch_layer_outputs, batch_classes)
        else:
            if layer_outputs is None:
                layer_outputs = batch_layer_outputs
            else:
                for i, (layer, batch_layer) in enumerate(zip(layer_outputs, batch_layer_outputs)):
                    layer_outputs[i] = np.concatenate((layer, batch_layer), axis=0)
            classes.extend(batch_classes)
    if use_cache:
        layer_outputs, classes = cache.load_and_transform(len(loader))
    else:
        fit = umap.UMAP(**umap_parameters)
        with Suppressor():  # Silence a bunch of numba warnings
            layer_outputs = [fit.fit_transform(layer) for layer in layer_outputs]
    draw_umaps(layer_outputs, classes, layer_ids=layers, layers=network.layers, save=save, save_path=save_dir,
               dictionary=load_dict(dictionary_path, True), legend_mode=legend_mode)



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import numpy as np
import tensorflow as tf

# Empirical (ImageNet) color correlation matrix, from https://github.com/tensorflow/lucid
# yapf: disable
color_correlation_svd_sqrt = np.asarray([[0.26, 0.09, 0.02],
                                         [0.27, 0.00, -0.05],
                                         [0.27, -0.09, 0.03]]).astype("float32")
# yapf: enable
max_norm_svd_sqrt = np.max(np.linalg.norm(color_correlation_svd_sqrt, axis=0))
color_correlation_normalized = color_correlation_svd_sqrt / max_norm_svd_sqrt
color_correlation_normalized_transpose = color_correlation_normalized.T

color_mean = [0.48, 0.46, 0.41]


@tf.function
def linear_decorelate_color(image):
    """Multiply input by sqrt of empirical color correlation matrix.

    If you interpret an image's innermost dimension as describing colors in a
    decorrelated version of the color space (which is a very natural way to
    describe colors -- see discussion in Feature Visualization article) the way
    to map back to normal colors is multiply the square root of your color
    correlations.
    """
    # TODO: check that inner dimension is 3
    t_flat = tf.reshape(image, [-1, 3])
    t_flat = tf.matmul(t_flat, color_correlation_normalized_transpose)
    image = tf.reshape(t_flat, tf.shape(image))
    return image


@tf.function
def to_valid_rgb(image, decorrelate=False, sigmoid=True):
    """Transform inner dimension of image to valid rgb colors.

    In practice this consists of two parts:
    (1) If requested, transform the colors from a decorrelated color space to RGB.
    (2) Constrain the color channels to be in [-1,1], either using a sigmoid
        function or clipping.
    Args:
        image: input tensor, innermost dimension will be interpreted as colors and transformed/constrained.
        decorrelate: should the input tensor's colors be interpreted as coming from a whitened space or not?
        sigmoid: should the colors be constrained using sigmoid (if True) or clipping (if False).
    Returns:
      'image' with the innermost dimension transformed.
    """
    # TODO let the user specify / infer the data range instead of forcing [-1,1]
    if decorrelate:
        image = linear_decorelate_color(image)
    if decorrelate and not sigmoid:
        image += color_mean
    if sigmoid:
        return 2.0 * tf.nn.sigmoid(image) - 1.0
    else:
        return tf.clip_by_value(image, -1.0, 1.0)



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import numpy as np
import tensorflow as tf
from tensorflow.python.keras import layers, models


def classification_sub_net(num_classes, num_anchor=9):
    model = models.Sequential()
    model.add(
        layers.Conv2D(256,
                      kernel_size=3,
                      strides=1,
                      padding='same',
                      activation='relu',
                      kernel_initializer=tf.random_normal_initializer(stddev=0.01)))
    model.add(
        layers.Conv2D(256,
                      kernel_size=3,
                      strides=1,
                      padding='same',
                      activation='relu',
                      kernel_initializer=tf.random_normal_initializer(stddev=0.01)))
    model.add(
        layers.Conv2D(256,
                      kernel_size=3,
                      strides=1,
                      padding='same',
                      activation='relu',
                      kernel_initializer=tf.random_normal_initializer(stddev=0.01)))
    model.add(
        layers.Conv2D(256,
                      kernel_size=3,
                      strides=1,
                      padding='same',
                      activation='relu',
                      kernel_initializer=tf.random_normal_initializer(stddev=0.01)))
    model.add(
        layers.Conv2D(num_classes * num_anchor,
                      kernel_size=3,
                      strides=1,
                      padding='same',
                      activation='sigmoid',
                      kernel_initializer=tf.random_normal_initializer(stddev=0.01),
                      bias_initializer=tf.initializers.constant(np.log(1 / 99))))
    model.add(layers.Reshape((-1, num_classes)))  # the output dimension is [batch, #anchor, #classes]
    return model


def regression_sub_net(num_anchor=9):
    model = models.Sequential()
    model.add(
        layers.Conv2D(256,
                      kernel_size=3,
                      strides=1,
                      padding='same',
                      activation='relu',
                      kernel_initializer=tf.random_normal_initializer(stddev=0.01)))
    model.add(
        layers.Conv2D(256,
                      kernel_size=3,
                      strides=1,
                      padding='same',
                      activation='relu',
                      kernel_initializer=tf.random_normal_initializer(stddev=0.01)))
    model.add(
        layers.Conv2D(256,
                      kernel_size=3,
                      strides=1,
                      padding='same',
                      activation='relu',
                      kernel_initializer=tf.random_normal_initializer(stddev=0.01)))
    model.add(
        layers.Conv2D(256,
                      kernel_size=3,
                      strides=1,
                      padding='same',
                      activation='relu',
                      kernel_initializer=tf.random_normal_initializer(stddev=0.01)))
    model.add(
        layers.Conv2D(4 * num_anchor,
                      kernel_size=3,
                      strides=1,
                      padding='same',
                      kernel_initializer=tf.random_normal_initializer(stddev=0.01)))
    model.add(layers.Reshape((-1, 4)))  # the output dimension is [batch, #anchor, 4]
    return model


def RetinaNet(input_shape, num_classes, num_anchor=9):
    inputs = tf.keras.Input(shape=input_shape)
    # FPN
    resnet50 = tf.keras.applications.ResNet50(weights="imagenet", include_top=False, input_tensor=inputs, pooling=None)
    assert resnet50.layers[80].name == "conv3_block4_out"
    C3 = resnet50.layers[80].output
    assert resnet50.layers[142].name == "conv4_block6_out"
    C4 = resnet50.layers[142].output
    assert resnet50.layers[-1].name == "conv5_block3_out"
    C5 = resnet50.layers[-1].output
    P5 = layers.Conv2D(256, kernel_size=1, strides=1, padding='same')(C5)
    P5_upsampling = layers.UpSampling2D()(P5)
    P4 = layers.Conv2D(256, kernel_size=1, strides=1, padding='same')(C4)
    P4 = layers.Add()([P5_upsampling, P4])
    P4_upsampling = layers.UpSampling2D()(P4)
    P3 = layers.Conv2D(256, kernel_size=1, strides=1, padding='same')(C3)
    P3 = layers.Add()([P4_upsampling, P3])
    P6 = layers.Conv2D(256, kernel_size=3, strides=2, padding='same', name="P6")(C5)
    P7 = layers.Activation('relu')(P6)
    P7 = layers.Conv2D(256, kernel_size=3, strides=2, padding='same', name="P7")(P7)
    P5 = layers.Conv2D(256, kernel_size=3, strides=1, padding='same', name="P5")(P5)
    P4 = layers.Conv2D(256, kernel_size=3, strides=1, padding='same', name="P4")(P4)
    P3 = layers.Conv2D(256, kernel_size=3, strides=1, padding='same', name="P3")(P3)
    # classification subnet
    cls_subnet = classification_sub_net(num_classes=num_classes, num_anchor=num_anchor)
    P3_cls = cls_subnet(P3)
    P4_cls = cls_subnet(P4)
    P5_cls = cls_subnet(P5)
    P6_cls = cls_subnet(P6)
    P7_cls = cls_subnet(P7)
    cls_output = layers.Concatenate(axis=-2)([P3_cls, P4_cls, P5_cls, P6_cls, P7_cls])
    # localization subnet
    loc_subnet = regression_sub_net(num_anchor=num_anchor)
    P3_loc = loc_subnet(P3)
    P4_loc = loc_subnet(P4)
    P5_loc = loc_subnet(P5)
    P6_loc = loc_subnet(P6)
    P7_loc = loc_subnet(P7)
    loc_output = layers.Concatenate(axis=-2)([P3_loc, P4_loc, P5_loc, P6_loc, P7_loc])
    return tf.keras.Model(inputs=inputs, outputs=[cls_output, loc_output])


def get_fpn_anchor_box(input_shape):
    assert len(input_shape) == 3
    h, w, _ = input_shape
    assert h % 32 == 0 and w % 32 == 0
    shapes = [(int(h / 8), int(w / 8))]  # P3
    num_pixel = np.prod(shapes)
    for _ in range(4):  # P4 through P7
        shapes.append((int(np.ceil(shapes[-1][0] / 2)), int(np.ceil(shapes[-1][1] / 2))))
        num_pixel += np.prod(shapes[-1])
    anchorbox = np.zeros((9 * num_pixel, 4))
    base_multipliers = [2**(0.0), 2**(1 / 3), 2**(2 / 3)]
    aspect_ratio_multiplier = [(1.0, 1.0), (2.0, 1.0), (1.0, 2.0)]
    anchor_idx = 0
    for shape in shapes:
        p_h, p_w = shape
        base_y = 1 / p_h
        base_x = 1 / p_w
        for i in range(p_h):
            for j in range(p_w):
                for base_multiplier in base_multipliers:
                    for aspect_x, aspect_y in aspect_ratio_multiplier:
                        center_y = (i + 1 / 2) * base_y
                        center_x = (j + 1 / 2) * base_x
                        anchorbox[anchor_idx, 0] = max(center_x - base_x * base_multiplier * aspect_x, 0.0)  # x1
                        anchorbox[anchor_idx, 1] = max(center_y - base_y * base_multiplier * aspect_y, 0.0)  # y1
                        anchorbox[anchor_idx, 2] = min(center_x + base_x * base_multiplier * aspect_x, 1.0)  # x2
                        anchorbox[anchor_idx, 3] = min(center_y + base_y * base_multiplier * aspect_y, 1.0)  # y2
                        anchor_idx += 1
        if p_h == 1 and p_w == 1:  # the next level of 1x1 feature map is still 1x1, therefore ignore
            break
    return np.float32(anchorbox)


def get_target(anchorbox, label, x1, y1, x2, y2, num_classes=10):
    num_anchor = anchorbox.shape[0]
    target_cls = np.zeros(shape=(num_anchor), dtype=np.int64)
    target_loc = np.zeros(shape=(num_anchor, 4), dtype=np.float32)
    for _label, _x1, _y1, _x2, _y2 in zip(label, x1, y1, x2, y2):
        best_iou = 0.0
        for anchor_idx in range(num_anchor):
            iou = get_iou((_x1, _y1, _x2, _y2), anchorbox[anchor_idx])
            if iou > best_iou:
                best_iou = iou
                best_anchor_idx = anchor_idx
            if iou > 0.5:
                target_cls[anchor_idx] = _label
                target_loc[anchor_idx] = get_loc_offset((_x1, _y1, _x2, _y2), anchorbox[anchor_idx])
            elif iou > 0.4:
                target_cls[anchor_idx] = -2  # ignore this example
            else:
                target_cls[anchor_idx] = -1  # background class
        if best_iou > 0 and best_iou < 0.5:  # if gt has no >0.5 iou with any anchor
            target_cls[best_anchor_idx] = _label
            target_loc[best_anchor_idx] = get_loc_offset((_x1, _y1, _x2, _y2), anchorbox[best_anchor_idx])
    return target_cls, target_loc


def get_loc_offset(box_gt, box_anchor):
    gt_x1, gt_y1, gt_x2, gt_y2 = tuple(box_gt)
    ac_x1, ac_y1, ac_x2, ac_y2 = tuple(box_anchor)
    anchor_width = ac_x2 - ac_x1
    anchor_height = ac_y2 - ac_y1
    dx1 = (gt_x1 - ac_x1) / anchor_width
    dy1 = (gt_y1 - ac_y1) / anchor_height
    dx2 = (gt_x2 - ac_x2) / anchor_width
    dy2 = (gt_y2 - ac_y2) / anchor_height
    return dx1, dy1, dx2, dy2


def get_iou(box1, box2):
    b1_x1, b1_y1, b1_x2, b1_y2 = tuple(box1)
    b2_x1, b2_y1, b2_x2, b2_y2 = tuple(box2)
    xA = max(b1_x1, b2_x1)
    yA = max(b1_y1, b2_y1)
    xB = min(b1_x2, b2_x2)
    yB = min(b1_y2, b2_y2)
    interArea = max(0, xB - xA) * max(0, yB - yA)
    if interArea == 0:
        iou = 0
    else:
        box1Area = (b1_x2 - b1_x1) * (b1_y2 - b1_y1)
        box2Area = (b2_x2 - b2_x1) * (b2_y2 - b2_y1)
        iou = interArea / (box1Area + box2Area - interArea)
    return iou



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import argparse
import os

from pyfiglet import Figlet


class SaveAction(argparse.Action):
    """
    A custom save action which is used to populate a secondary variable inside of an exclusive group. Used if this file
        is invoked directly during argument parsing.
    """
    def __init__(self, option_strings, dest, nargs='?', **kwargs):
        if '?' != nargs:
            raise ValueError("nargs must be \'?\'")
        super().__init__(option_strings, dest, nargs, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, True)
        setattr(namespace, self.dest + '_dir', values if values is None else os.path.join(values, ''))


def draw():
    print(Figlet(font="slant").renderText("FastEstimator"))



import tensorflow as tf


class ReflectionPadding2D(tf.keras.layers.Layer):
    def __init__(self, padding=(1, 1)):
        self.padding = tuple(padding)
        self.input_spec = [tf.keras.layers.InputSpec(ndim=4)]
        super().__init__()

    def get_config(self):
        return {'padding': self.padding}

    def compute_output_shape(self, s):
        return (s[0], s[1] + 2 * self.padding[0], s[2] + 2 * self.padding[1], s[3])

    def call(self, x, mask=None):
        w_pad, h_pad = self.padding
        return tf.pad(x, [[0, 0], [h_pad, h_pad], [w_pad, w_pad], [0, 0]], 'REFLECT')


class InstanceNormalization(tf.keras.layers.Layer):
    """Instance Normalization Layer (https://arxiv.org/abs/1607.08022)."""
    def __init__(self, epsilon=1e-5):
        super().__init__()
        self.epsilon = epsilon

    def get_config(self):
        return {'epsilon': self.epsilon}

    def build(self, input_shape):
        self.scale = self.add_weight(name='scale',
                                     shape=input_shape[-1:],
                                     initializer=tf.random_normal_initializer(0., 0.02),
                                     trainable=True)

        self.offset = self.add_weight(name='offset', shape=input_shape[-1:], initializer='zeros', trainable=True)

    def call(self, x):
        mean, variance = tf.nn.moments(x, axes=[1, 2], keepdims=True)
        inv = tf.math.rsqrt(variance + self.epsilon)
        normalized = (x - mean) * inv
        return self.scale * normalized + self.offset


def _residual_block(x0, num_filter, kernel_size=(3, 3), strides=(1, 1)):
    initializer = tf.random_normal_initializer(0., 0.02)
    x0_cropped = tf.keras.layers.Cropping2D(cropping=2)(x0)

    x = tf.keras.layers.Conv2D(filters=num_filter,
                               kernel_size=kernel_size,
                               strides=strides,
                               kernel_initializer=initializer)(x0)
    x = InstanceNormalization()(x)
    x = tf.keras.layers.ReLU()(x)

    x = tf.keras.layers.Conv2D(filters=num_filter,
                               kernel_size=kernel_size,
                               strides=strides,
                               kernel_initializer=initializer)(x)

    x = InstanceNormalization()(x)
    x = tf.keras.layers.Add()([x, x0_cropped])
    return x


def _conv_block(x0, num_filter, kernel_size=(9, 9), strides=(1, 1), padding="same", apply_relu=True):
    initializer = tf.random_normal_initializer(0., 0.02)
    x = tf.keras.layers.Conv2D(filters=num_filter,
                               kernel_size=kernel_size,
                               strides=strides,
                               padding=padding,
                               kernel_initializer=initializer)(x0)

    x = InstanceNormalization()(x)
    if apply_relu:
        x = tf.keras.layers.ReLU()(x)
    return x


def _upsample(x0, num_filter, kernel_size=(3, 3), strides=(2, 2), padding="same"):
    initializer = tf.random_normal_initializer(0., 0.02)
    x = tf.keras.layers.Conv2DTranspose(filters=num_filter,
                                        kernel_size=kernel_size,
                                        strides=strides,
                                        padding=padding,
                                        kernel_initializer=initializer)(x0)

    x = InstanceNormalization()(x)
    x = tf.keras.layers.ReLU()(x)
    return x


def _downsample(x0, num_filter, kernel_size=(3, 3), strides=(2, 2), padding="same"):
    initializer = tf.random_normal_initializer(0., 0.02)
    x = tf.keras.layers.Conv2D(filters=num_filter,
                               kernel_size=kernel_size,
                               strides=strides,
                               padding=padding,
                               kernel_initializer=initializer)(x0)

    x = InstanceNormalization()(x)
    x = tf.keras.layers.ReLU()(x)
    return x


def styleTransferNet(input_shape=(256, 256, 3), num_resblock=5):
    x0 = tf.keras.layers.Input(shape=input_shape)
    x = ReflectionPadding2D(padding=(40, 40))(x0)
    x = _conv_block(x, num_filter=32)
    x = _downsample(x, num_filter=64)
    x = _downsample(x, num_filter=128)

    for _ in range(num_resblock):
        x = _residual_block(x, num_filter=128)

    x = _upsample(x, num_filter=64)
    x = _upsample(x, num_filter=32)
    x = _conv_block(x, num_filter=3, apply_relu=False)
    x = tf.keras.layers.Activation("tanh")(x)
    return tf.keras.Model(inputs=x0, outputs=x)


def lossNet(input_shape=(256, 256, 3),
            stlyeLayers=["block1_conv2", "block2_conv2", "block3_conv3", "block4_conv3"],
            contentLayers=["block3_conv3"]):
    x0 = tf.keras.layers.Input(shape=input_shape)
    mdl = tf.keras.applications.vgg16.VGG16(include_top=False, weights='imagenet', input_tensor=x0)
    # Compute style loss
    styleOutput = [mdl.get_layer(name).output for name in stlyeLayers]
    contentOutput = [mdl.get_layer(name).output for name in contentLayers]
    output = {"style": styleOutput, "content": contentOutput}
    return tf.keras.Model(inputs=x0, outputs=output)



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import os

import matplotlib.pyplot as plt
import numpy as np
# noinspection PyPackageRequirements
import tensorflow as tf
# noinspection PyPackageRequirements
from tensorflow.python import keras

from fastestimator.util.loader import PathLoader
from fastestimator.util.saliencies import GradientSaliency, IntegratedGradients
from fastestimator.util.util import is_number, load_dict, load_image, decode_predictions
from fastestimator.util.vis_util import show_image, show_text, show_gray_image


@tf.function
def compute_percentile(tensor, percentile, keepdims=True):
    """
    Args:
        tensor: A tensor with batches on the zero axis. Shape (batch X ...)
        percentile: The percentile value to be computed for each batch in the tensor
        keepdims: Whether to keep shape compatibility with the input tensor
    Returns:
        A tensor corresponding to the given percentile value within each batch of the input tensor
    """
    result = tf.reduce_min(
        tf.math.top_k(tf.reshape(tensor, (tensor.shape[0], -1)),
                      tf.cast(tf.math.ceil((1 - percentile / 100) * tensor.shape[1] * tensor.shape[2]), tf.int32),
                      sorted=False).values, axis=1)
    if keepdims:
        result = tf.reshape(result, [tensor.shape[0]] + [1 for _ in tensor.shape[1:]])
    return result


@tf.function
def convert_for_visualization(batched_masks, percentile=99):
    """
    Args:
        batched_masks: Input masks, channel values to be reduced by absolute value summation
        percentile: The percentile [0-100] used to set the max value of the image
    Returns:
        A (batch X width X height) image after visualization clipping is applied
    """
    flattened_mask = tf.reduce_sum(tf.abs(batched_masks), axis=3)

    vmax = compute_percentile(flattened_mask, percentile)
    vmin = tf.reduce_min(flattened_mask, axis=(1, 2), keepdims=True)

    return tf.clip_by_value((flattened_mask - vmin) / (vmax - vmin), 0, 1)


def interpret_model(model, model_input, baseline_input=None, decode_dictionary=None, color_map="inferno", smooth=7,
                    save=False, save_path='.'):
    """ Displays or saves a saliency mask interpretation of the given input
    Args:
        model: A model to evaluate. Should be a classifier which takes the 0th axis as the batch axis
        model_input: Input tensor, shaped for the model ex. (1, 299, 299, 3)
        baseline_input: An example of what a blank model input would be.
                        Should be a tensor with the same shape as model_input
        decode_dictionary: A dictionary of "class_idx" -> "class_name" associations
        color_map: The color map to use to visualize the saliency maps.
                        Consider "Greys_r", "plasma", or "magma" as alternatives
        smooth: The number of samples to use when generating a smoothed image
        save: Whether to save (True) or display (False) the result
        save_path: Where to save the output if save=True
    """
    predictions = np.asarray(model(model_input))
    decoded = decode_predictions(predictions, top=3, dictionary=decode_dictionary)

    grad_sal = GradientSaliency(model)
    grad_int = IntegratedGradients(model)

    vanilla_masks = grad_sal.get_mask(model_input)
    vanilla_ims = convert_for_visualization(vanilla_masks)
    smooth_masks = grad_sal.get_smoothed_mask(model_input, nsamples=smooth)
    smooth_ims = convert_for_visualization(smooth_masks)
    smooth_integrated_masks = grad_int.get_smoothed_mask(model_input, nsamples=smooth, input_baseline=baseline_input)
    smooth_integrated_ims = convert_for_visualization(smooth_integrated_masks)

    filtered_inputs = (model_input + 1) * np.asarray(smooth_integrated_ims)[:, :, :, None] - 1

    num_rows = model_input.shape[0]
    num_cols = 6
    dpi = 96.0

    box_width = max(220, model_input.shape[2])
    box_height = max(220, model_input.shape[1])
    fig, axs = plt.subplots(num_rows, num_cols, figsize=(num_cols * (box_width / dpi), num_rows * (box_height / dpi)),
                            dpi=dpi)
    if num_rows == 1:
        axs = [axs]  # axs object not wrapped if there's only one row
    for i in range(num_rows):
        show_text(axs[i][0], np.ones_like(model_input[i]), decoded[i], title="Predictions" if i == 0 else None)
        show_image(axs[i][1], model_input[i], title="Raw" if i == 0 else None)
        show_image(axs[i][2], filtered_inputs[i], title="Filtered" if i == 0 else None)
        show_gray_image(axs[i][3], vanilla_ims[i], color_map=color_map, title="Vanilla" if i == 0 else None)
        show_gray_image(axs[i][4], smooth_ims[i], color_map=color_map, title="Smoothed" if i == 0 else None)
        show_gray_image(axs[i][5], smooth_integrated_ims[i], color_map=color_map,
                        title="Integrated Smoothed" if i == 0 else None)
    plt.subplots_adjust(top=0.95, bottom=0.01, left=0.01, right=0.99, hspace=0.03, wspace=0.03)
    # plt.tight_layout(pad=0.3, h_pad=0.03, w_pad=0.03, rect=(0, 0, 0.98, 0.98))

    if not save:
        plt.show()
    else:
        if save_path is None or save_path == "":
            save_path = "."
        os.makedirs(save_path, exist_ok=True)
        save_file = os.path.join(save_path, 'saliency.png')
        print("Saving to %s" % save_file)
        plt.savefig(save_file, dpi=300, bbox_inches="tight")


def load_and_interpret(model_path, input_paths, baseline=-1, dictionary_path=None, strip_alpha=False, smooth_factor=7,
                       save=False, save_dir=None):
    """ A helper class to load input and invoke the interpretation api

    Args:
        model_path: The path the model file (str)
        input_paths: The paths to model input files [(str),...] or to a folder of inputs [(str)]
        baseline: Either a number corresponding to the baseline for integration, or a path to a baseline file
        dictionary_path: The path to a dictionary file encoding a 'class_idx'->'class_name' mapping
        strip_alpha: Whether to collapse alpha channels when loading an input (bool)
        smooth_factor: How many iterations of the smoothing algorithm to run (int)
        save: Whether to save (True) or display (False) the resulting image
        save_dir: Where to save the image if save=True
    """
    model_dir = os.path.dirname(model_path)
    if save_dir is None:
        save_dir = model_dir
    network = keras.models.load_model(model_path)
    input_type = network.input.dtype
    input_shape = network.input.shape
    n_channels = 0 if len(input_shape) == 3 else input_shape[3]

    dic = load_dict(dictionary_path)
    if len(input_paths) == 1 and os.path.isdir(input_paths[0]):
        loader = PathLoader(input_paths[0])
        input_paths = [path[0] for path in loader.path_pairs]
    inputs = [load_image(input_paths[i], strip_alpha=strip_alpha, channels=n_channels) for i in range(len(input_paths))]
    max_shapes = np.maximum.reduce([inp.shape for inp in inputs], axis=0)
    tf_image = tf.stack([
        tf.image.resize_with_crop_or_pad(tf.convert_to_tensor(im, dtype=input_type), max_shapes[0], max_shapes[1])
        for im in inputs
    ], axis=0)
    if is_number(baseline):
        baseline_gen = tf.constant_initializer(float(baseline))
        baseline_image = baseline_gen(shape=tf_image.shape, dtype=input_type)
    else:
        baseline_image = load_image(baseline)
        baseline_image = tf.convert_to_tensor(baseline_image, dtype=input_type)

    interpret_model(network, tf_image, baseline_input=baseline_image, decode_dictionary=dic, smooth=smooth_factor,
                    save=save, save_path=save_dir)



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from fastestimator.util.op import TensorOp


class Scheduler:
    def __init__(self, epoch_dict):
        self.epoch_dict = epoch_dict
        self.value = None
        self._verify_inputs()

    def _verify_inputs(self):
        assert isinstance(self.epoch_dict, dict), "must provide dictionary as epoch_dict"
        self.keys = sorted(self.epoch_dict)
        sample_content = self._get_sample_content()
        for key in self.keys:
            assert isinstance(key, int), "found non-integer key: {}".format(key)
            assert key >= 0, "found negative key: {}".format(key)
            if isinstance(sample_content, TensorOp) and self.epoch_dict[key]:
                assert self.mode == self.epoch_dict[key].mode, "schedule contents must have same mode"

    def get_sequential_value(self, epoch):
        if epoch in self.keys:
            self.value = self.epoch_dict[epoch]
        return self.value

    def get_current_value(self, epoch):
        if epoch in self.keys:
            value = self.epoch_dict[epoch]
        else:
            last_key = self._get_last_key(epoch)
            if last_key is None:
                value = None
            else:
                value = self.epoch_dict[last_key]
        return value

    def _get_last_key(self, epoch):
        last_key = None
        for key in self.keys:
            if key > epoch:
                break
            last_key = key
        return last_key

    def _get_sample_content(self):
        sample_content = None
        for value in self.epoch_dict.values():
            if value is not None:
                sample_content = value
                break
        if isinstance(sample_content, TensorOp):
            self.mode = sample_content.mode
        assert sample_content is not None, "At least one value in a scheduler dict must be non-None"
        return sample_content



import math

import numpy as np


class LRSchedule:
    def __init__(self, schedule_mode):
        self.schedule_mode = schedule_mode
        self.total_epochs = None  #will be filled by runtime
        self.total_train_steps = None  #will be filled by runtime
        self.initial_lr = None  #will be filled by runtime

    def schedule_fn(self, current_step_or_epoch, lr):
        #do something to change lr
        return lr


class CyclicLRSchedule(LRSchedule):
    def __init__(self, num_cycle=1, cycle_multiplier=2, decrease_method="cosine"):
        super().__init__(schedule_mode="step")
        self.num_cycle = num_cycle
        self.cycle_multiplier = cycle_multiplier
        self.decrease_method = decrease_method
        self.decay_fn_map = {"cosine": self.lr_cosine_decay, "linear": self.lr_linear_decay}
        assert self.cycle_multiplier >= 1, "The cycle_multiplier should at least be 1"

    def lr_linear_decay(self, current_step, lr_start, lr_end, step_start, step_end):
        slope = (lr_start - lr_end) / (step_start - step_end)
        intercept = lr_start - slope * step_start
        lr = slope * current_step + intercept
        return np.float32(lr)

    def lr_cosine_decay(self, current_step, lr_start, lr_end, step_start, step_end):
        current_step_in_cycle = (current_step - step_start) / (step_end - step_start)
        lr = (lr_start - lr_end) / 2 * math.cos(current_step_in_cycle * math.pi) + (lr_start + lr_end) / 2
        return np.float32(lr)

    def schedule_fn(self, current_step_or_epoch, lr):
        if self.cycle_multiplier == 1:
            total_unit_cycles = self.num_cycle
        else:
            total_unit_cycles = (self.cycle_multiplier**self.num_cycle - 1) / (self.cycle_multiplier - 1)
        unit_cycle_length = self.total_train_steps // total_unit_cycles
        step_start = 0
        for i in range(self.num_cycle):
            current_cycle_length = unit_cycle_length * self.cycle_multiplier**i
            if (current_step_or_epoch - step_start) < current_cycle_length:
                step_end = step_start + current_cycle_length
                break
            else:
                if i == (self.num_cycle - 1):
                    step_end = self.total_train_steps
                else:
                    step_start = step_start + current_cycle_length
        lr = self.decay_fn_map[self.decrease_method](current_step_or_epoch, self.initial_lr, 1e-6, step_start, step_end)
        return np.float32(lr)



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import os
import tarfile
import tempfile
from operator import add

import h5py
import pandas as pd
import wget


def get_name(index, hdf5_data):
    name = hdf5_data['/digitStruct/name']
    return ''.join([chr(v[0]) for v in hdf5_data[name[index][0]].value])


def get_bbox(index, hdf5_data):
    attrs = {}
    item = hdf5_data['digitStruct']['bbox'][index].item()
    for key in ['label', 'left', 'top', 'width', 'height']:
        attr = hdf5_data[item][key]
        values = [int(hdf5_data[attr.value[i].item()].value[0][0])
                  for i in range(len(attr))] if len(attr) > 1 else [int(attr.value[0][0])]
        attrs[key] = values
    return attrs


def img_boundingbox_data_constructor(data_folder, mode, csv_path):
    f = h5py.File(os.path.join(data_folder, "digitStruct.mat"), 'r')
    row_list = []
    num_example = f['/digitStruct/bbox'].shape[0]
    logging_interval = num_example // 10
    print("found %d number of examples for %s" % (num_example, mode))
    for j in range(num_example):
        if j % logging_interval == 0:
            print("retrieving bounding box for %s: %.2f%%" % (mode, j / num_example * 100))
        img_name = get_name(j, f)
        bbox = get_bbox(j, f)
        row_dict = {
            'image': os.path.join(mode, img_name),
            'label': bbox["label"],
            'x1': bbox["left"],
            'y1': bbox["top"],
            'x2': list(map(add, bbox["left"], bbox["width"])),
            'y2': list(map(add, bbox["top"], bbox["height"]))
        }
        row_list.append(row_dict)
    bbox_df = pd.DataFrame(row_list, columns=['image', 'label', 'x1', 'y1', 'x2', 'y2'])
    bbox_df.to_csv(csv_path, index=False)
    return bbox_df


def load_data(path=None):
    if path is None:
        path = os.path.join(tempfile.gettempdir(), ".fe", "SVHN")
    if not os.path.exists(path):
        os.makedirs(path)
    train_csv = os.path.join(path, "train_data.csv")
    test_csv = os.path.join(path, "test_data.csv")
    train_compressed_path = os.path.join(path, "train.tar.gz")
    test_compressed_path = os.path.join(path, "test.tar.gz")
    train_folder_path = os.path.join(path, "train")
    test_folder_path = os.path.join(path, "test")
    if not (os.path.exists(train_compressed_path) and os.path.exists(test_compressed_path)):
        print("Downloading data to %s" % path)
        wget.download('http://ufldl.stanford.edu/housenumbers/train.tar.gz', path)
        wget.download("http://ufldl.stanford.edu/housenumbers/test.tar.gz", path)
    if not (os.path.exists(train_folder_path) and os.path.exists(test_folder_path)):
        print(" ")
        print("extracting data...")
        with tarfile.open(train_compressed_path) as tar:
            tar.extractall(path)
        with tarfile.open(test_compressed_path) as tar:
            tar.extractall(path)
    if not (os.path.exists(train_csv) and os.path.exists(test_csv)):
        print("constructing bounding box data...")
        img_boundingbox_data_constructor(train_folder_path, "train", train_csv)
        img_boundingbox_data_constructor(test_folder_path, "test", test_csv)
    return train_csv, test_csv, path



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import os

import numpy as np
from scipy.io import loadmat

from fastestimator.util.op import NumpyOp

EPSILON = 1e-7


class ImageReader(NumpyOp):
    """
    Class for reading png or jpg images

    Args:
        parent_path: Parent path that will be added on given path
        grey_scale: Boolean to indicate whether or not to read image as grayscale
    """
    def __init__(self, inputs=None, outputs=None, mode=None, parent_path="", grey_scale=False):
        super().__init__(inputs=inputs, outputs=outputs, mode=mode)
        import cv2
        self.parent_path = parent_path
        self.color_flag = cv2.IMREAD_COLOR
        if grey_scale:
            self.color_flag = cv2.IMREAD_GRAYSCALE
        self.forward_fn = cv2.imread

    def forward(self, path, state):
        """
        Reads numpy array from image path

        Args:
            path: path of the image
            state: A dictionary containing background information such as 'mode'

        Returns:
           Image as numpy array
        """
        path = os.path.normpath(os.path.join(self.parent_path, path))
        data = self.forward_fn(path, self.color_flag)
        if not isinstance(data, np.ndarray):
            raise ValueError('cv2 did not read correctly for file "{}"'.format(path))
        return data


class Zscore(NumpyOp):
    """
    Standardize data using zscore method
    """
    def forward(self, data, state):
        """
        Standardizes the data

        Args:
            data: Data to be standardized
            state: A dictionary containing background information such as 'mode'

        Returns:
            Array containing standardized data
        """
        mean = np.mean(data)
        std = np.std(data)
        data = (data - mean) / max(std, EPSILON)
        return data


class Minmax(NumpyOp):
    """
    Normalize data using the minmax method
    """
    def forward(self, data, state):
        """
        Normalizes the data

        Args:
            data: Data to be normalized
            state: A dictionary containing background information such as 'mode'

        Returns:
            Normalized numpy array
        """
        data_max = np.max(data)
        data_min = np.min(data)
        data = (data - data_min) / max((data_max - data_min), EPSILON)
        return data


class Scale(NumpyOp):
    """
    Preprocessing class for scaling dataset

    Args:
        scalar: Scalar for scaling the data
    """
    def __init__(self, scalar, inputs=None, outputs=None, mode=None):
        super().__init__(inputs=inputs, outputs=outputs, mode=mode)
        self.scalar = scalar

    def forward(self, data, state):
        """
        Scales the data tensor

        Args:
            data: Data to be scaled
            state: A dictionary containing background information such as 'mode'

        Returns:
            Scaled data array
        """
        data = self.scalar * data
        return data


class Reshape(NumpyOp):
    """
    Preprocessing class for reshaping the data

    Args:
        shape: target shape
    """
    def __init__(self, shape, inputs=None, outputs=None, mode=None):
        super().__init__(inputs=inputs, outputs=outputs, mode=mode)
        self.shape = shape

    def forward(self, data, state):
        """
        Reshapes data array

        Args:
            data: Data to be reshaped
            state: A dictionary containing background information such as 'mode'

        Returns:
            Reshaped array
        """
        data = np.reshape(data, self.shape)
        return data


class MatReader(NumpyOp):
    """Class for reading .mat files.

    Args:
        parent_path: Parent path that will be added on given path.
    """
    def __init__(self, inputs=None, outputs=None, mode=None, parent_path=""):
        super().__init__(inputs=inputs, outputs=outputs, mode=mode)
        self._loadmat = loadmat
        self.parent_path = parent_path

    def forward(self, data, state):
        """Reads mat file as dict.

        Args:
            data: Path to the mat file.
            state: A dictionary containing background information such as 'mode'

        Returns:
           dict
        """
        path = os.path.normpath(os.path.join(self.parent_path, data))
        data = self._loadmat(path)
        return data


class Resize(NumpyOp):
    """Resize image.

    Args:
        target_size (tuple): Target image size in (height, width) format.
        resize_method (string): `bilinear`, `nearest`, `area`, and `lanczos4` are available.
        keep_ratio (bool): If `True`, the resulting image will be padded to keep the original aspect ratio.

    Returns:
        Resized `np.ndarray`.
    """
    def __init__(self, target_size, resize_method='bilinear', keep_ratio=False, inputs=None, outputs=None, mode=None):
        super().__init__(inputs=inputs, outputs=outputs, mode=mode)
        import cv2
        self._cv2 = cv2
        self.target_size = target_size
        if resize_method == "bilinear":
            self.resize_method = cv2.INTER_LINEAR
        elif resize_method == "nearest":
            self.resize_method = cv2.INTER_NEAREST
        elif resize_method == "area":
            self.resize_method = cv2.INTER_AREA
        elif resize_method == "lanczos4":
            self.resize_method = cv2.INTER_LANCZOS4
        self.keep_ratio = keep_ratio

    def forward(self, data, state):
        if self.keep_ratio:
            original_ratio = data.shape[1] / data.shape[0]
            target_ratio = self.target_size[1] / self.target_size[0]
            if original_ratio >= target_ratio:
                pad = (data.shape[1] / target_ratio - data.shape[0]) / 2
                pad_boarder = (np.ceil(pad).astype(np.int), np.floor(pad).astype(np.int), 0, 0)
            else:
                pad = (data.shape[0] * target_ratio - data.shape[1]) / 2
                pad_boarder = (0, 0, np.ceil(pad).astype(np.int), np.floor(pad).astype(np.int))
            data = self._cv2.copyMakeBorder(data, *pad_boarder, self._cv2.BORDER_CONSTANT)
        data = self._cv2.resize(data, (self.target_size[1], self.target_size[0]), self.resize_method)
        return data



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import tensorflow as tf

from fastestimator.util.op import TensorOp

EPSILON = 1e-7


class TensorFilter(TensorOp):
    def __init__(self, inputs, mode="train"):
        super().__init__(inputs=inputs, mode=mode)

    def forward(self, data, state):
        return tf.constant(True)


class ScalarFilter(TensorFilter):
    """Class for performing filtering on dataset based on scalar values.

    Args:
        inputs: Name of the key in the dataset that is to be filtered
        filter_value: The values in the dataset that are to be filtered.
        keep_prob: The probability of keeping the example
        mode: mode that the filter acts on
    """
    def __init__(self, inputs, filter_value, keep_prob, mode="train"):
        super().__init__(inputs=inputs, mode=mode)
        self.filter_value = filter_value
        self.keep_prob = keep_prob
        self._verify_inputs()

    def _verify_inputs(self):
        assert isinstance(self.inputs, str), "ScalarFilter only accepts single string input"
        self.filter_value = ScalarFilter._convert_to_list(self.filter_value)
        self.keep_prob = ScalarFilter._convert_to_list(self.keep_prob)
        assert len(self.filter_value) == len(self.keep_prob), "filter_value and keep_prob must be same length"

    @staticmethod
    def _convert_to_list(data):
        if not isinstance(data, list):
            if isinstance(data, tuple):
                data = list(data)
            else:
                data = [data]
        return data

    def forward(self, data, state):
        pass_filter = tf.constant(True)
        for filter_value, keep_prob in zip(self.filter_value, self.keep_prob):
            pass_current_filter = tf.cond(tf.equal(data, filter_value),
                                          lambda: tf.greater(keep_prob, tf.random.uniform([])),
                                          lambda: tf.constant(True))
            pass_filter = tf.logical_and(pass_filter, pass_current_filter)
        return pass_filter


class Binarize(TensorOp):
    """
    Binarize data based on threshold between 0 and 1

    Args:
        threshold: Threshold for binarizing
    """
    def __init__(self, threshold, inputs=None, outputs=None, mode=None):
        super().__init__(inputs=inputs, outputs=outputs, mode=mode)
        self.thresh = threshold

    def forward(self, data, state):
        """
        Transforms the image to binary based on threshold

        Args:
            data: Data to be binarized
            state: Information about the current execution context

        Returns:
            Tensor containing binarized data
        """
        data = tf.math.greater(data, self.thresh)
        data = tf.cast(data, tf.float32)
        return data


class Zscore(TensorOp):
    """
    Standardize data using zscore method
    """
    def forward(self, data, state):
        """
        Standardizes the data tensor

        Args:
            data: Data to be standardized
            state: Information about the current execution context

        Returns:
            Tensor containing standardized data
        """
        data = tf.cast(data, tf.float32)
        mean = tf.reduce_mean(data)
        std = tf.keras.backend.std(data)
        data = tf.math.divide(tf.subtract(data, mean), tf.maximum(std, EPSILON))
        data = tf.cast(data, tf.float32)
        return data


class Minmax(TensorOp):
    """
    Normalize data using the minmax method
    """
    def forward(self, data, state):
        """
        Normalizes the data tensor

        Args:
            data: Data to be normalized
            state: Information about the current execution context

        Returns:
            Tensor after minmax
        """
        data = tf.cast(data, tf.float32)
        data = tf.math.divide(tf.subtract(data, tf.reduce_min(data)),
                              tf.maximum(tf.subtract(tf.reduce_max(data), tf.reduce_min(data)), EPSILON))
        return data


class Scale(TensorOp):
    """
    Preprocessing class for scaling dataset

    Args:
        scalar: Scalar for scaling the data
    """
    def __init__(self, scalar, inputs=None, outputs=None, mode=None):
        super().__init__(inputs=inputs, outputs=outputs, mode=mode)
        self.scalar = scalar

    def forward(self, data, state):
        """
        Scales the data tensor

        Args:
            data: Data to be scaled
            state: Information about the current execution context

        Returns:
            Scaled data tensor
        """
        data = tf.cast(data, tf.float32)
        data = tf.scalar_mul(self.scalar, data)
        return data


class Onehot(TensorOp):
    """
    Preprocessing class for converting categorical labels to onehot encoding

    Args:
        num_dim: Number of dimensions of the labels
    """
    def __init__(self, num_dim, inputs=None, outputs=None, mode=None):
        super().__init__(inputs=inputs, outputs=outputs, mode=mode)
        self.num_dim = num_dim

    def forward(self, data, state):
        """
        Transforms categorical labels to onehot encodings

        Args:
            data: Data to be preprocessed
            state: Information about the current execution context

        Returns:
            Transformed labels
        """
        data = tf.cast(data, tf.int32)
        data = tf.one_hot(data, self.num_dim)
        return data


class Resize(TensorOp):
    """
    Preprocessing class for resizing the images

    Args:
        size: Destination shape of the images
        resize_method: One of resize methods provided by tensorflow to be used
    """
    def __init__(self, size, resize_method=tf.image.ResizeMethod.BILINEAR, inputs=None, outputs=None, mode=None):
        super().__init__(inputs=inputs, outputs=outputs, mode=mode)
        self.size = size
        self.resize_method = resize_method

    def forward(self, data, state):
        """
        Resizes data tensor

        Args:
            data: Tensor to be resized
            state: Information about the current execution context

        Returns:
            Resized tensor
        """
        preprocessed_data = tf.image.resize(data, self.size, self.resize_method)
        return preprocessed_data


class Reshape(TensorOp):
    """
    Preprocessing class for reshaping the data

    Args:
        shape: target shape
    """
    def __init__(self, shape, inputs=None, outputs=None, mode=None):
        super().__init__(inputs=inputs, outputs=outputs, mode=mode)
        self.shape = shape

    def forward(self, data, state):
        """
        Reshapes data tensor

        Args:
            data: Data to be reshaped
            state: Information about the current execution context

        Returns:
            Reshaped tensor
        """
        data = tf.reshape(data, self.shape)
        return data



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import argparse
import sys

from fastestimator.util.cli_util import SaveAction
from fastestimator.util.util import parse_cli_to_dictionary
from fastestimator.visualization.caricature import load_and_caricature
from fastestimator.visualization.layer_umap import umap_layers
from fastestimator.visualization.parse_logs import parse_folder
from fastestimator.visualization.saliency_masks import load_and_interpret


def logs(args, unknown):
    if len(unknown) > 0:
        print("error: unrecognized arguments: ", str.join(", ", unknown))
        sys.exit(-1)
    parse_folder(args['log_dir'], args['extension'], args['recursive'], args['smooth'], args['save'], args['save_dir'],
                 args['ignore'], args['share_legend'], args['pretty_names'])


def configure_log_parser(subparsers):
    parser = subparsers.add_parser('logs', description='Generates comparison graphs amongst one or more log files',
                                   formatter_class=argparse.ArgumentDefaultsHelpFormatter, allow_abbrev=False)
    parser.add_argument('log_dir', metavar='<Log Dir>', type=str,
                        help="The path to a folder containing one or more log files")
    parser.add_argument('--extension', metavar='E', type=str, help="The file type / extension of your logs",
                        default=".txt")
    parser.add_argument('--recursive', action='store_true', help="Recursively search sub-directories for log files")
    parser.add_argument('--ignore', metavar='I', type=str, nargs='+',
                        help="The names of metrics to ignore though they may be present in the log files")
    parser.add_argument('--smooth', metavar='<float>', type=float,
                        help="The amount of gaussian smoothing to apply (zero for no smoothing)", default=1)
    parser.add_argument('--pretty_names', help="Clean up the metric names for display", action='store_true')

    legend_group = parser.add_argument_group('legend arguments')
    legend_x_group = legend_group.add_mutually_exclusive_group(required=False)
    legend_x_group.add_argument('--common_legend', dest='share_legend', help="Generate one legend total",
                                action='store_true', default=True)
    legend_x_group.add_argument('--split_legend', dest='share_legend', help="Generate one legend per graph",
                                action='store_false', default=False)

    save_group = parser.add_argument_group('output arguments')
    save_x_group = save_group.add_mutually_exclusive_group(required=False)
    save_x_group.add_argument(
        '--save', nargs='?', metavar='<Save Dir>', dest='save', action=SaveAction, default=False,
        help="Save the output image. May be accompanied by a directory into \
                                              which the file is saved. If no output directory is specified, the log \
                                              directory will be used")
    save_x_group.add_argument('--display', dest='save', action='store_false',
                              help="Render the image to the UI (rather than saving it)", default=True)
    save_x_group.set_defaults(save_dir=None)
    parser.set_defaults(func=logs)


def saliency(args, unknown):
    if len(unknown) > 0:
        print("error: unrecognized arguments: ", str.join(", ", unknown))
        sys.exit(-1)
    load_and_interpret(args['model'], args['inputs'], baseline=args['baseline'], dictionary_path=args['dictionary'],
                       strip_alpha=args['strip_alpha'], smooth_factor=args['smooth'], save=args['save'],
                       save_dir=args['save_dir'])


def configure_saliency_parser(subparsers):
    parser = subparsers.add_parser('saliency', description='Generates saliency maps for a model on given input(s)',
                                   formatter_class=argparse.ArgumentDefaultsHelpFormatter, allow_abbrev=False)
    parser.add_argument('model', metavar='<Model Path>', type=str, help="The path to a saved model file")
    parser.add_argument(
        'inputs', metavar='Input', type=str, nargs='+',
        help="The paths to one or more inputs to visualize (or the path to a folder which will be \
                        recursively traversed to find image files)")
    parser.add_argument('--dictionary', metavar='<Dictionary Path>', type=str,
                        help="The path to a {'class_id':'class_name'} json file", default=None)
    parser.add_argument('--smooth', metavar='N', type=int, default=7,
                        help="The number of samples to use when generating smoothed saliency masks")
    parser.add_argument(
        '--baseline', metavar='B', type=str, default='-1',
        help="The value to use as a baseline for integrated gradient calculations. Can be either a \
                        number or the path to an image. What would a 'blank' input look like")
    parser.add_argument('--strip-alpha', action='store_true', help="True if you want to convert RGBA images to RGB")
    save_group = parser.add_argument_group('output arguments')
    save_x_group = save_group.add_mutually_exclusive_group(required=False)
    save_x_group.add_argument(
        '--save', nargs='?', metavar='<Save Dir>',
        help="Save the output image. May be accompanied by a directory into which the \
                              file is saved. If no output directory is specified, the model directory will be used",
        dest='save', action=SaveAction, default=False)
    save_x_group.add_argument('--display', dest='save', action='store_false',
                              help="Render the image to the UI (rather than saving it)", default=True)
    save_x_group.set_defaults(save_dir=None)
    parser.set_defaults(func=saliency)


def umap(args, unknown):
    hyperparameters = parse_cli_to_dictionary(unknown)
    umap_layers(args['model'], args['input_dir'], print_layers=args['print_layers'], strip_alpha=args['strip_alpha'],
                layers=args['layers'], batch=args['batch'], use_cache=args['cache'], cache_dir=args['cache_dir'],
                dictionary_path=args['dictionary'], legend_mode=args['legend'], save=args['save'],
                save_dir=args['save_dir'], umap_parameters=hyperparameters, input_extension=args['extension'])


def configure_umap_parser(subparsers):
    parser = subparsers.add_parser('umap', description='Plots umaps for model layers over given inputs',
                                   formatter_class=argparse.ArgumentDefaultsHelpFormatter, allow_abbrev=False)
    parser.add_argument('model', metavar='<Model Path>', type=str, help="The path to a saved model file")
    parser.add_argument('input_dir', metavar='Input', type=str,
                        help="The path to a folder of folders containing inputs sorted by class")
    parser.add_argument(
        '--print-layers', action='store_true',
        help="True if you only want to print a model's layers to the console. Useful for inspecting a \
                            model to choose which layer indices to visualize.")
    parser.add_argument(
        '--layers', metavar='IDX', type=int, nargs='+',
        help='The indices of layers you want to visualize. If not provided then all layers will be \
                                      analyzed, but that is likely to run out of memory on moderately sized models.')
    parser.add_argument('--strip-alpha', action='store_true', help="True if you want to convert RGBA images to RGB")
    parser.add_argument('--extension', metavar='E', type=str,
                        help="The file type / extension of your input files (default matches all files)", default=None)
    parser.add_argument('--batch', type=int, default=50,
                        help="How many images to process at a time (set based on available RAM)")
    parser.add_argument('--dictionary', metavar='<Dictionary Path>', type=str,
                        help="The path to a {'class_id':'class_name'} json file", default=None)
    parser.add_argument('--legend', choices=['on', 'off', 'shared'], default='shared',
                        help='Whether to display legends or not.')
    save_group = parser.add_argument_group('output arguments')
    save_group.add_argument(
        '--cache', metavar='<Cache Dir>', action=SaveAction, default=False,
        help="Enable caching of intermediate results. Highly recommended to reduce RAM \
                            requirements. If no cache directory is specified, a sibling of the input directory will be \
                            used")
    save_group.set_defaults(cache_dir=None)
    save_x_group = save_group.add_mutually_exclusive_group(required=False)
    save_x_group.add_argument(
        '--save', nargs='?', metavar='<Save Dir>', dest='save', action=SaveAction, default=False,
        help="Save the output image. May be accompanied by a directory into \
                                                      which the file is saved. If no output directory is specified, \
                                                      the model directory will be used")
    save_x_group.add_argument('--display', dest='save', action='store_false',
                              help="Render the image to the UI (rather than saving it)", default=True)
    save_x_group.set_defaults(save_dir=None)
    parser.add_argument_group(
        'umap arguments', 'Arguments from the umap module may all be passed through. Examples \
    include --n_neighbors <int>, --metric <str>, --n_epochs <int>, --min_dist <float>, --learning_rate <float>, etc...')
    parser.set_defaults(func=umap)


def caricature(args, unknown):
    if len(unknown) > 0:
        print("error: unrecognized arguments: ", str.join(", ", unknown))
        sys.exit(-1)
    load_and_caricature(args['model'], args['inputs'], dictionary_path=args['dictionary'], save=args['save'],
                        save_dir=args['save_dir'], strip_alpha=args['strip_alpha'], layer_ids=args['layers'],
                        print_layers=args['print_layers'], n_steps=args['steps'], learning_rate=args['learning_rate'],
                        blur=args['blur'], cossim_pow=args['cossim'], sd=args['stddev'], fft=not args['pixel_based'],
                        decorrelate=not args['true_color'], sigmoid=not args['hard_clip'])


def configure_caricature_parser(subparsers):
    parser = subparsers.add_parser('caricature', description='Plots caricatures for model layers over given inputs',
                                   formatter_class=argparse.ArgumentDefaultsHelpFormatter, allow_abbrev=False)
    parser.add_argument('model', metavar='<Model Path>', type=str, help="The path to a saved model file")
    parser.add_argument(
        'inputs', metavar='Input', type=str, nargs='+',
        help="The paths to one or more inputs to visualize (or the path to a folder which will be \
                            recursively traversed to find image files)")
    parser.add_argument(
        '--print-layers', action='store_true',
        help="True if you only want to print a model's layers to the console. Useful for inspecting a \
                            model to choose which layer indices to visualize.")
    parser.add_argument(
        '--layers', metavar='IDX', type=int, nargs='+',
        help='The indices of layers you want to visualize. If not provided then all layers will be \
                                      analyzed, but that is likely to run out of memory on moderately sized models.')
    parser.add_argument('--strip-alpha', action='store_true', help="True if you want to convert RGBA images to RGB")
    parser.add_argument('--dictionary', metavar='<Dictionary Path>', type=str,
                        help="The path to a {'class_id':'class_name'} json file", default=None)
    parser.add_argument('--steps', metavar='N', type=int, help="How many optimization iterations to run", default=250)
    parser.add_argument('--learning-rate', metavar='LR', type=float, help="Initial learning rate for the optimizer",
                        default=0.05)
    parser.add_argument('--blur', metavar='B', type=float,
                        help="Strength of blurring used on image during processing (reduces high-frequency artifacts)",
                        default=1.0)
    parser.add_argument(
        '--cossim', metavar='B', type=float,
        help="Strength of cosine similarity coefficient in loss function (higher value increases \
                        similarity)", default=0.5)
    parser.add_argument('--stddev', type=float,
                        help="The standard deviation of the noise used to generate the seed image", default=0.01)
    parser.add_argument(
        '--pixel-based', action='store_true',
        help="Use pixel-based optimization rather than fft based. Individual steps will be faster but \
                        convergence will be slower")
    parser.add_argument('--true-color', action='store_true',
                        help="Disable color decorrelation. You may want this if using non-imagenet style images")
    parser.add_argument('--hard-clip', action='store_true',
                        help="Use hard clipping instead of sigmoids to constrain pixel values")
    save_group = parser.add_argument_group('output arguments')
    save_x_group = save_group.add_mutually_exclusive_group(required=False)
    save_x_group.add_argument(
        '--save', nargs='?', metavar='<Save Dir>',
        help="Save the output image. May be accompanied by a directory into which the \
                              file is saved. If no output directory is specified, the model directory will be used",
        dest='save', action=SaveAction, default=False)
    save_x_group.add_argument('--display', dest='save', action='store_false',
                              help="Render the image to the UI (rather than saving it)", default=True)
    save_x_group.set_defaults(save_dir=None)
    parser.set_defaults(func=caricature)


def configure_visualization_parser(subparsers):
    visualization_parser = subparsers.add_parser('visualize', description='Generates various types of visualiztaions',
                                                 allow_abbrev=False)
    visualization_subparsers = visualization_parser.add_subparsers()
    # In python 3.7 the following 2 lines could be put into the .add_subparsers() call
    visualization_subparsers.required = True
    visualization_subparsers.dest = 'mode'

    configure_log_parser(visualization_subparsers)
    configure_saliency_parser(visualization_subparsers)
    configure_umap_parser(visualization_subparsers)
    configure_caricature_parser(visualization_subparsers)



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""
Download montgomery  dataset from  http://openi.nlm.nih.gov/imgs/collections/NLM-MontgomeryCXRSet.zip
"""
import os
import tempfile
from glob import glob
import zipfile
import shutil

import numpy as np
import pandas as pd
import wget

def load_data(path=None):
    if path is None:
        path = os.path.join(tempfile.gettempdir(), ".fe", 'MONTGOMERY')
    if not os.path.exists(path):
        os.makedirs(path)
    csv_path = os.path.join(path, "montgomery.csv")
    data_compressed_path = os.path.join(path, 'NLM-MontgomeryCXRSet.zip')
    extract_folder_path = os.path.join(path, 'MontgomerySet')
    if not os.path.exists(data_compressed_path):
        print("Downloading data...")
        wget.download('http://openi.nlm.nih.gov/imgs/collections/NLM-MontgomeryCXRSet.zip', path)
    if not os.path.exists(extract_folder_path):
        print(" ")
        print("Extracting data...")
        with zipfile.ZipFile(data_compressed_path, 'r') as zip_file:
            zip_file.extractall(path)
        shutil.move(os.path.join(extract_folder_path, "ManualMask", "leftMask"), os.path.join(extract_folder_path, "leftMask"))
        shutil.move(os.path.join(extract_folder_path, "ManualMask", "rightMask"), os.path.join(extract_folder_path, "rightMask"))
    if not os.path.exists(csv_path):
        img_list = glob(os.path.join(extract_folder_path, "CXR_png", '*.png'))
        df = pd.DataFrame(data={'image': img_list})
        df['image'] = df['image'].apply(lambda x: os.path.relpath(x, path))
        df['mask_left'] = df['image'].str.replace('CXR_png', 'leftMask')
        df['mask_right'] = df['image'].str.replace('CXR_png', 'rightMask')
        df.to_csv(csv_path, index=False)
    return csv_path, path



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
import tensorflow as tf
import umap
from matplotlib.lines import Line2D

from fastestimator.estimator.trace import Trace
from fastestimator.util.util import Suppressor


class UMap(Trace):
    def __init__(self,
                 in_vector_key,
                 label_vector_key=None,
                 label_dict=None,
                 output_name=None,
                 legend_loc='best',
                 im_freq=1,
                 **umap_parameters):
        """
        Args:
            in_vector_key: The key of the input to be fed into the umap algorithm
            label_vector_key: The (optional) key of the classes corresponding to the inputs (used for coloring points)
            label_dict: An (optional) dictionary mapping labels from the label vector to other representations 
                        (ex. {0:'dog', 1:'cat'})
            output_name: The key which the umap image will be saved into within the state dictionary
            legend_loc: The location of the legend, or 'off' to disable figure legends
            im_freq: Frequency (in epochs) during which visualizations should be generated
            **umap_parameters: Extra parameters to be passed to the umap algorithm, ex. n_neighbors, n_epochs, etc. 
        """
        if output_name is None:
            output_name = "{}_umap".format(in_vector_key)
        super().__init__(inputs={in_vector_key, label_vector_key}, outputs=output_name, mode='eval')
        self.in_key = in_vector_key
        self.label_key = label_vector_key
        self.output_key = output_name
        self.fit = umap.UMAP(**umap_parameters)
        self.n_components = umap_parameters.get("n_components", 2)
        self.data = []
        self.labels = []
        self.color_dict = None
        self.label_dict = label_dict
        self.legend_elems = None
        self.legend_loc = legend_loc
        self.im_freq = im_freq
        self.recording = False

    def on_epoch_begin(self, state):
        if state['epoch'] % self.im_freq == 0:
            self.recording = True
        else:
            self.recording = False

    def on_batch_end(self, state):
        if self.recording:
            self.data.append(state.get(self.in_key) or state['batch'][self.in_key])
            if self.label_key:
                self.labels.append(state.get(self.label_key) or state['batch'][self.label_key])

    def on_epoch_end(self, state):
        if not self.recording:
            return
        color_list = self._map_classes_to_colors(tf.concat(self.labels, axis=0))
        if self.legend_elems is None and color_list is not None:
            self.legend_elems = [
                Line2D([0], [0],
                       marker='o',
                       color='w',
                       markerfacecolor=self.color_dict[clazz],
                       label=clazz if self.label_dict is None else self.label_dict[clazz],
                       markersize=7) for clazz in self.color_dict
            ]
        with Suppressor():  # Silence a bunch of numba warnings
            points = self.fit.fit_transform(tf.concat(self.data, axis=0))
        fig = plt.figure()
        ax = fig.add_subplot(111)
        ax.set_yticks([], [])
        ax.set_yticklabels([])
        ax.set_xticks([], [])
        ax.set_xticklabels([])
        if self.n_components == 1:
            ax.scatter(points[:, 0], range(len(points)), c=color_list or 'b', s=3)
        if self.n_components == 2:
            ax.scatter(points[:, 0], points[:, 1], c=color_list or 'b', s=3)
        if self.n_components == 3:
            ax.scatter(points[:, 0], points[:, 1], points[:, 2], c=color_list or 'b', s=3)
        if self.legend_elems and self.legend_loc != 'off':
            ax.legend(handles=self.legend_elems, loc=self.legend_loc, fontsize='small')
        plt.tight_layout()
        # TODO - Figure out how to get this to work without it displaying the figure. maybe fig.canvas.draw
        plt.draw()
        plt.pause(0.000001)
        state.maps[1][self.output_key] = np.fromstring(fig.canvas.tostring_rgb(), dtype=np.uint8,
                                                       sep='').reshape((1, ) + fig.canvas.get_width_height()[::-1] +
                                                                       (3, ))
        plt.close(fig)
        self.data.clear()
        self.labels.clear()

    def _map_classes_to_colors(self, classifications):
        if classifications is None or len(classifications) == 0:
            return None
        if self.color_dict is None:
            classes = set(map(lambda x: int(x), classifications))
            num_classes = len(classes)
            colors = sns.hls_palette(n_colors=num_classes,
                                     s=0.95) if num_classes > 10 else sns.color_palette("colorblind")
            class_to_color = {clazz: idx for idx, clazz in enumerate(classes)}
            self.color_dict = {clazz: colors[class_to_color[clazz]] for clazz in classes}
        return [self.color_dict[int(clazz)] for clazz in classifications]



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import json
import multiprocessing as mp
import os
import time

import numpy as np
import tensorflow as tf

from fastestimator.pipeline.processing import TensorFilter
from fastestimator.record.record import RecordWriter
from fastestimator.util.op import get_inputs_by_key, get_inputs_by_op, get_op_from_mode, verify_ops, \
    write_outputs_by_key
from fastestimator.util.schedule import Scheduler
from fastestimator.util.tfrecord import get_features
from fastestimator.util.util import convert_tf_dtype, flatten_list, get_num_devices


class Pipeline:
    def __init__(self,
                 data,
                 batch_size,
                 ops=None,
                 read_feature=None,
                 padded_batch=False,
                 expand_dims=False,
                 max_shuffle_buffer_mb=3000):

        self.batch_size = batch_size
        self.data = data
        self.ops = ops
        self.read_feature = read_feature
        self.padded_batch = padded_batch
        self.expand_dims = expand_dims
        self.max_shuffle_buffer_mb = max_shuffle_buffer_mb
        self.possible_mode = ["train", "eval"]
        self.padded_shape = None
        self.global_batch_multiplier = 1
        self.num_core = mp.cpu_count()
        self._verify_input()
        self._reset()

    def _verify_input(self):
        assert isinstance(self.data, (dict, RecordWriter, str)), \
            "data must be either RecordWriter, dictionary or record path"
        if self.read_feature:
            assert isinstance(self.read_feature,
                              (list, tuple, dict)), "read_feature must be either list, tuple or dictionary"
            if not isinstance(self.read_feature, tuple):
                self.read_feature = [self.read_feature]
        if self.ops:
            if not isinstance(self.ops, list):
                self.ops = [self.ops]
        else:
            self.ops = []

    def _reset(self):
        self.mode_list = []
        self.all_features = {"train": [], "eval": []}
        self.num_examples = {"train": [], "eval": []}
        self.shuffle_buffer = {"train": [], "eval": []}
        self.feature_name = {"train": [], "eval": []}
        self.extracted_dataset = {}
        self.transformed_dataset = {}
        self.dataset_schedule = {}
        self.all_output_keys = set()
        # TFrecord only
        self.summary_file = {}
        self.feature_dtype = {"train": [], "eval": []}
        self.record_feature_shape = {"train": [], "eval": []}
        self.compression = {"train": [], "eval": []}
        self.file_names = {"train": [], "eval": []}
        self.global_batch_multiplier = 1

    def prepare(self, distribute_strategy=None):
        if isinstance(self.data, dict):
            self._get_numpy_config()
        elif isinstance(self.data, (RecordWriter, str)):
            if isinstance(self.data, RecordWriter):
                data_path = self.data.save_dir
                if not os.path.exists(data_path) or not os.listdir(data_path):
                    self.data.write()
            else:
                data_path = self.data
            print("FastEstimator: Reading non-empty directory: {}".format(data_path))
            self._get_tfrecord_config(data_path)
        else:
            raise ValueError("data must be one of the following: dictionary, RecordWriter or record path")
        self.all_output_keys = self.all_output_keys | set(flatten_list(list(self.feature_name.values())))
        for mode in self.mode_list:
            self._get_feature_name(mode)
            self._extract_dataset(mode)
            self._transform_dataset(mode, distribute_strategy)
        self.all_output_keys = self.all_output_keys - {None}

    def _get_numpy_config(self):
        for mode in self.possible_mode:
            if mode in self.data:
                self.mode_list.append(mode)
                self.all_features[mode].append(self.data[mode])
                self._get_numpy_config_mode(mode)

    def _get_numpy_config_mode(self, mode):
        num_examples_list = []
        data = self.data[mode]
        for key in data.keys():
            feature_data = data[key]
            if type(feature_data) is list:
                num_examples_list.append(len(feature_data))
            elif type(feature_data) is np.ndarray:
                num_examples_list.append(feature_data.shape[0])
            else:
                raise ValueError("the feature only supports list or numpy array")
        assert len(set(
            num_examples_list)) == 1, "inconsistent number of data found during {}, please check the data".format(mode)
        self.num_examples[mode].append(set(num_examples_list).pop())
        self.shuffle_buffer[mode].append(set(num_examples_list).pop())

    def _get_tfrecord_config(self, data_path):
        found_data = False
        for mode in self.possible_mode:
            self.summary_file[mode] = [
                os.path.join(data_path, f) for f in os.listdir(data_path)
                if f.endswith(".json") and f.startswith("%s_summary" % mode)
            ]
            if len(self.summary_file[mode]) > 0:
                self.mode_list.append(mode)
                self._get_tfrecord_config_mode(data_path, mode)
                found_data = True
        assert found_data, "could not find data summary file in {}".format(data_path)

    def _get_tfrecord_config_mode(self, data_path, mode):
        for json_file in self.summary_file[mode]:
            with open(json_file, 'r') as fp:
                summary = json.load(fp)
            file_names = [os.path.join(data_path, f) for f in summary["file_names"]]
            self.file_names[mode].append(file_names)
            num_examples = np.sum(summary["num_examples"])
            example_size_mb = summary["example_size_mb"]
            self.num_examples[mode].append(num_examples)
            self.feature_dtype[mode].append(summary["feature_dtype"])
            self.record_feature_shape[mode].append(summary["feature_shape"])
            if "compression" in summary:
                compression = summary["compression"]
            else:
                compression = None
            self.compression[mode].append(compression)
            self.all_features[mode].append(get_features(file_names[0], compression=compression))
            self.shuffle_buffer[mode].append(int(min(num_examples, self.max_shuffle_buffer_mb // example_size_mb)))
            print("FastEstimator: Found %d examples for %s in %s" % (int(num_examples), mode, json_file))

    def _get_feature_name(self, mode):
        if len(self.all_features[mode]) > 1 and self.read_feature:
            assert isinstance(self.read_feature, tuple), "read feature must be a tuple for unpaired feature set"
            assert len(self.read_feature) == len(
                self.all_features[mode]), "the tuple should be consistent between read_feature and data"
        for idx, feature in enumerate(self.all_features[mode]):
            if self.read_feature is None:
                self.feature_name[mode].append(list(feature.keys()))
            elif isinstance(self.read_feature[idx], dict):
                self.feature_name[mode].append(self.read_feature[idx][mode])
            else:
                self.feature_name[mode].append(self.read_feature[idx])

    def _extract_dataset(self, mode):
        ds_tuple = ()
        # Data Reading
        for idx in range(len(self.all_features[mode])):
            if isinstance(self.data, dict):
                ds = tf.data.Dataset.from_tensor_slices(self.all_features[mode][idx])
            else:
                if mode == "train":
                    ds = tf.data.Dataset.from_tensor_slices(self.file_names[mode][idx])
                    ds = ds.shuffle(len(self.file_names[mode][idx]))
                    ds = ds.interleave(
                        lambda ds_lam: tf.data.TFRecordDataset(ds_lam, compression_type=self.compression[mode][idx]),
                        cycle_length=self.num_core,
                        block_length=2)
                else:
                    ds = tf.data.TFRecordDataset(self.file_names[mode][idx],
                                                 compression_type=self.compression[mode][idx])
                ds = ds.map(lambda ds_lam: self._decode_records(ds_lam, mode, idx), num_parallel_calls=self.num_core)
            ds = ds.shuffle(self.shuffle_buffer[mode][idx])
            ds = ds.repeat()
            ds_tuple += ds,
        # Combine dataset from different unpaired feature sets
        if len(self.all_features[mode]) > 1:
            dataset = tf.data.Dataset.zip(ds_tuple)
            dataset = dataset.map(self._combine_dataset, num_parallel_calls=self.num_core)
        else:
            dataset = ds_tuple[0]
        self.extracted_dataset[mode] = dataset

    def _transform_dataset(self, mode, distribute_strategy):
        all_output_keys = []
        signature_epoch, mode_ops = self._get_signature_epoch(mode)
        extracted_ds = self.extracted_dataset[mode]
        state = {"mode": mode}
        dataset_map = {}
        for epoch in signature_epoch:
            epoch_ops_all = []
            forward_ops_epoch = []
            filter_ops_epoch = []
            forward_ops_between_filter = []
            # get batch size for the epoch
            global_batch_size = self.get_global_batch_size(epoch)
            # generate ops for specific mode and epoch
            for op in mode_ops:
                if isinstance(op, Scheduler):
                    scheduled_op = op.get_current_value(epoch)
                    if scheduled_op:
                        epoch_ops_all.append(scheduled_op)
                else:
                    epoch_ops_all.append(op)
            # check the ops
            epoch_ops_without_filter = [op for op in epoch_ops_all if not isinstance(op, TensorFilter)]
            verify_ops(epoch_ops_without_filter, "Pipeline")
            # arrange operation according to filter location
            for op in epoch_ops_all:
                all_output_keys.append(op.outputs)
                if not isinstance(op, TensorFilter):
                    forward_ops_between_filter.append(op)
                else:
                    forward_ops_epoch.append(forward_ops_between_filter)
                    filter_ops_epoch.append(op)
                    forward_ops_between_filter = []
            forward_ops_epoch.append(forward_ops_between_filter)
            # execute the operations
            dataset = self._execute_ops(extracted_ds, forward_ops_epoch, filter_ops_epoch, state)
            # rest of the dataset setup
            if self.expand_dims:
                dataset = dataset.flat_map(tf.data.Dataset.from_tensor_slices)
            if self.padded_batch:
                _ = dataset.map(self._get_padded_shape)
                dataset = dataset.padded_batch(global_batch_size, padded_shapes=self.padded_shape)
            else:
                dataset = dataset.batch(global_batch_size)
            dataset = dataset.prefetch(buffer_size=1)
            if distribute_strategy:
                dataset = distribute_strategy.experimental_distribute_dataset(dataset)
            dataset_map[epoch] = iter(dataset)
        self.dataset_schedule[mode] = Scheduler(epoch_dict=dataset_map)
        self.all_output_keys = self.all_output_keys | set(flatten_list(all_output_keys))

    def _get_padded_shape(self, dataset):
        padded_shape = {}
        for key in dataset:
            padded_shape[key] = dataset[key].shape
        self.padded_shape = padded_shape
        return dataset

    def _execute_ops(self, dataset, forward_ops_epoch, filter_ops_epoch, state):
        num_filters = len(filter_ops_epoch)
        forward_ops = forward_ops_epoch[0]
        dataset = dataset.map(lambda ds: self._preprocess_fn(ds, forward_ops, state), num_parallel_calls=self.num_core)
        if num_filters > 0:
            for filter_op, forward_ops in zip(filter_ops_epoch, forward_ops_epoch[1:]):
                dataset = dataset.filter(lambda ds: self._filter_fn(ds, filter_op, state))
                dataset = dataset.map(lambda ds: self._preprocess_fn(ds, forward_ops, state),
                                      num_parallel_calls=self.num_core)
        return dataset

    @staticmethod
    def _filter_fn(feature, filter_op, state):
        data = get_inputs_by_key(feature, filter_op.inputs)
        data = filter_op.forward(data, state)
        return data

    @staticmethod
    def _preprocess_fn(feature, forward_ops, state):
        data = None
        for op in forward_ops:
            data = get_inputs_by_op(op, feature, data)
            data = op.forward(data, state)
            if op.outputs:
                feature = write_outputs_by_key(feature, data, op.outputs)
        return feature

    def _get_signature_epoch(self, mode):
        signature_epoch = [0]
        if isinstance(self.batch_size, Scheduler):
            signature_epoch.extend(self.batch_size.keys)
        mode_ops = get_op_from_mode(self.ops, mode)
        for op in mode_ops:
            if isinstance(op, Scheduler):
                signature_epoch.extend(op.keys)
        return list(set(signature_epoch)), mode_ops

    def _decode_records(self, dataset, mode, idx):
        decoded_data = {}
        all_data = tf.io.parse_single_example(dataset, features=self.all_features[mode][idx])
        for feature in self.feature_name[mode][idx]:
            data = all_data[feature]
            if "str" in str(data.dtype) and "str" not in self.feature_dtype[mode][idx][feature]:
                data = tf.io.decode_raw(data, convert_tf_dtype(self.feature_dtype[mode][idx][feature]))
                data = tf.reshape(data, self.record_feature_shape[mode][idx][feature])
            if "int" in str(data.dtype):
                data = tf.cast(data, tf.int32)
            elif "str" not in self.feature_dtype[mode][idx][feature]:
                data = tf.cast(data, tf.float32)
            decoded_data[feature] = data
        return decoded_data

    @staticmethod
    def _combine_dataset(*dataset):
        combined_dict = {}
        for ds in dataset:
            for key in ds.keys():
                assert key not in combined_dict, "found duplicated key {} in unpaird feature sets".format(key)
                combined_dict[key] = ds[key]
        return combined_dict

    def get_global_batch_size(self, epoch):
        batch_per_device = self.batch_size
        if isinstance(batch_per_device, Scheduler):
            batch_per_device = batch_per_device.get_current_value(epoch)
        global_batch_size = batch_per_device * self.global_batch_multiplier
        return global_batch_size

    def show_results(self, mode="train", num_steps=1, current_epoch=0):
        data = []
        self.global_batch_multiplier = get_num_devices()
        self.prepare()
        ds_iter = self.dataset_schedule[mode].get_current_value(current_epoch)
        for _ in range(num_steps):
            data.append(next(ds_iter))
        self._reset()
        return data

    def benchmark(self, mode="train", num_steps=1000, log_interval=100, current_epoch=0):
        self.global_batch_multiplier = get_num_devices()
        global_batch_size = self.get_global_batch_size(current_epoch)
        self.prepare()
        ds_iter = self.dataset_schedule[mode].get_current_value(current_epoch)
        start = time.perf_counter()
        for idx in range(num_steps + 1):
            _ = next(ds_iter)
            if idx % log_interval == 0:
                if idx == 0:
                    start = time.perf_counter()
                else:
                    duration = time.perf_counter() - start
                    example_per_sec = log_interval * global_batch_size / duration
                    print("FastEstimator: Step: %d, Epoch: %d, Batch Size %d, Example/sec %.2f" %
                          (idx, current_epoch, global_batch_size, example_per_sec))
                    start = time.perf_counter()
        self._reset()
